import logging

from netqasm.backend.messages import deserialize_host_msg as deserialize_message
from netqasm.backend.network_stack import CREATE_FIELDS
from netqasm.backend.network_stack import OK_FIELDS_K as OK_FIELDS
from netqasm.lang import instr as instructions
from netqasm.lang.encoding import RegisterName
from netqasm.lang.operand import Address, ArrayEntry, ArraySlice, Immediate, Register
from netqasm.lang.parsing import deserialize as deserialize_subroutine
from netqasm.lang.parsing.text import parse_text_subroutine
from netqasm.lang.subroutine import Subroutine
from netqasm.lang.version import NETQASM_VERSION
from netqasm.logging.glob import set_log_level
from netqasm.qlink_compat import EPRType, TimeUnit
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.epr_socket import EPRSocket
from netqasm.sdk.qubit import Qubit

DebugConnection.node_ids = {
    "Alice": 0,
    "Bob": 1,
}


def test_simple():
    set_log_level(logging.DEBUG)
    with DebugConnection("Alice") as alice:
        q1 = Qubit(alice)
        q2 = Qubit(alice)
        q1.H()
        q2.X()
        q1.X()
        q2.H()

    # 4 messages: init, subroutine, stop app and stop backend
    assert len(alice.storage) == 4
    raw_subroutine = deserialize_message(raw=alice.storage[1]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)
    expected = Subroutine(
        netqasm_version=NETQASM_VERSION,
        app_id=0,
        instructions=[
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(0),
            ),
            instructions.core.QAllocInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.InitInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(1),
            ),
            instructions.core.QAllocInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.InitInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(0),
            ),
            instructions.vanilla.GateHInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(1),
            ),
            instructions.vanilla.GateXInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(0),
            ),
            instructions.vanilla.GateXInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(1),
            ),
            instructions.vanilla.GateHInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            # NOTE qubits are now freed when application ends
            # without explicit qfree for each
            # Command(instruction=Instruction.SET, operands=[
            #     Register(RegisterName.Q, 0),
            #     0,
            # ]),
            # Command(instruction=Instruction.QFREE, operands=[
            #     Register(RegisterName.Q, 0),
            # ]),
            # Command(instruction=Instruction.SET, operands=[
            #     Register(RegisterName.Q, 0),
            #     1,
            # ]),
            # Command(instruction=Instruction.QFREE, operands=[
            #     Register(RegisterName.Q, 0),
            # ]),
        ],
    )
    for instr, expected_instr in zip(subroutine.instructions, expected.instructions):
        print(repr(instr))
        print(repr(expected_instr))
        assert instr == expected_instr
    print(subroutine)
    print(expected)


def test_rotations():
    set_log_level(logging.DEBUG)
    with DebugConnection("Alice") as alice:
        q = Qubit(alice)
        q.rot_X(n=1, d=1)

    # 4 messages: init, subroutine, stop app and stop backend
    assert len(alice.storage) == 4
    raw_subroutine = deserialize_message(raw=alice.storage[1]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)
    expected = Subroutine(
        netqasm_version=NETQASM_VERSION,
        app_id=0,
        instructions=[
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(0),
            ),
            instructions.core.QAllocInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.InitInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(0),
            ),
            instructions.vanilla.RotXInstruction(
                reg=Register(RegisterName.Q, 0),
                imm0=Immediate(1),
                imm1=Immediate(1),
            ),
        ],
    )
    for instr, expected_instr in zip(subroutine.instructions, expected.instructions):
        print(repr(instr))
        print(repr(expected_instr))
        assert instr == expected_instr
    print(subroutine)
    print(expected)


def test_epr_k_create():

    set_log_level(logging.DEBUG)

    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        q1 = epr_socket.create_keep()[0]
        q1.H()

    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
    assert len(alice.storage) == 5
    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)
    print(subroutine)

    expected_text = """
# NETQASM 0.0
# APPID 0
set R0 10
array R0 @0
set R0 1
array R0 @1
set R0 0
set R1 0
store R0 @1[R1]
set R0 20
array R0 @2
set R0 0
set R1 0
store R0 @2[R1]
set R0 1
set R1 1
store R0 @2[R1]
set R0 1
set R1 0
set R2 1
set R3 2
set R4 0
create_epr R0 R1 R2 R3 R4
set R0 0
set R1 10
wait_all @0[R0:R1]
set Q0 0
h Q0
ret_arr @0
ret_arr @1
ret_arr @2
"""

    expected = parse_text_subroutine(expected_text)

    for i, instr in enumerate(subroutine.instructions):
        print(repr(instr))
        expected_instr = expected.instructions[i]
        print(repr(expected_instr))
        print()
        assert instr == expected_instr
    print(subroutine)
    print(expected)


def test_epr_k_recv():

    set_log_level(logging.DEBUG)

    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        q1 = epr_socket.recv_keep()[0]
        q1.H()

    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
    assert len(alice.storage) == 5
    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)
    print(subroutine)

    expected_text = """
# NETQASM 0.0
# APPID 0
set R5 10
array R5 @0
set R5 1
array R5 @1
set R5 0
set R6 0
store R5 @1[R6]
set R5 1
set R6 0
set R7 1
set R8 0
recv_epr R5 R6 R7 R8
set R5 0
set R6 10
wait_all @0[R5:R6]
set R2 0
set R5 1
beq R2 R5 42
load R0 @1[R2]
set R3 9
set R4 0
beq R4 R2 27
set R5 10
add R3 R3 R5
set R5 1
add R4 R4 R5
jmp 21
load R1 @0[R3]
set R0 0
set R5 3
bne R1 R5 32
rot_z R0 16 4
set R5 1
bne R1 R5 35
rot_x R0 16 4
set R5 2
bne R1 R5 39
rot_x R0 16 4
rot_z R0 16 4
set R5 1
add R2 R2 R5
jmp 16
set Q0 0
h Q0
ret_arr @0
ret_arr @1
ret_arr @2
"""

    expected = parse_text_subroutine(expected_text)

    for i, instr in enumerate(subroutine.instructions):
        print(repr(instr))
        expected_instr = expected.instructions[i]
        print(repr(expected_instr))
        print()
        assert instr == expected_instr
    print(subroutine)
    print(expected)


def test_two_epr_k_create():

    set_log_level(logging.DEBUG)

    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        qubits = epr_socket.create(number=2)
        qubits[0].H()
        qubits[1].H()

    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
    assert len(alice.storage) == 5
    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)
    print(subroutine)

    expected_text = """
# NETQASM 0.0
# APPID 0
set R0 20
array R0 @0
set R0 2
array R0 @1
set R0 0
set R1 0
store R0 @1[R1]
set R0 1
set R1 1
store R0 @1[R1]
set R0 20
array R0 @2
set R0 0
set R1 0
store R0 @2[R1]
set R0 2
set R1 1
store R0 @2[R1]
set R0 1
set R1 0
set R2 1
set R3 2
set R4 0
create_epr R0 R1 R2 R3 R4
set R0 0
set R1 20
wait_all @0[R0:R1]
set Q0 0
h Q0
set Q0 1
h Q0
ret_arr @0
ret_arr @1
ret_arr @2
    """

    expected = parse_text_subroutine(expected_text)

    for i, instr in enumerate(subroutine.instructions):
        print(repr(instr))
        expected_instr = expected.instructions[i]
        print(repr(expected_instr))
        assert instr == expected_instr
    print(subroutine)
    print(expected)


def test_two_epr_k_recv():

    set_log_level(logging.DEBUG)

    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        qubits = epr_socket.recv(number=2)
        qubits[0].H()
        qubits[1].H()

    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
    assert len(alice.storage) == 5
    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)
    print(subroutine)

    expected_text = """
# NETQASM 0.0
# APPID 0
set R5 20
array R5 @0
set R5 2
array R5 @1
set R5 0
set R6 0
store R5 @1[R6]
set R5 1
set R6 1
store R5 @1[R6]
set R5 1
set R6 0
set R7 1
set R8 0
recv_epr R5 R6 R7 R8
set R5 0
set R6 20
wait_all @0[R5:R6]
set R2 0
set R5 2
beq R2 R5 45
load R0 @1[R2]
set R3 9
set R4 0
beq R4 R2 30
set R5 10
add R3 R3 R5
set R5 1
add R4 R4 R5
jmp 24
load R1 @0[R3]
set R0 0
set R5 3
bne R1 R5 35
rot_z R0 16 4
set R5 1
bne R1 R5 38
rot_x R0 16 4
set R5 2
bne R1 R5 42
rot_x R0 16 4
rot_z R0 16 4
set R5 1
add R2 R2 R5
jmp 19
set Q0 0
h Q0
set Q0 1
h Q0
ret_arr @0
ret_arr @1
ret_arr @2
    """

    expected = parse_text_subroutine(expected_text)

    for i, instr in enumerate(subroutine.instructions):
        print(repr(instr))
        expected_instr = expected.instructions[i]
        print(repr(expected_instr))
        assert instr == expected_instr
    print(subroutine)
    print(expected)


def test_epr_m():

    set_log_level(logging.DEBUG)

    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        outcomes = epr_socket.create_measure()
        m = outcomes[0].raw_measurement_outcome
        with m.if_eq(0):
            m.add(1)

    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
    assert len(alice.storage) == 5
    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)
    print(subroutine)
    expected = Subroutine(
        netqasm_version=NETQASM_VERSION,
        app_id=0,
        instructions=[
            # Arg array
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(OK_FIELDS),
            ),
            instructions.core.ArrayInstruction(
                reg=Register(RegisterName.R, 1),
                address=Address(0),
            ),
            # ent info array
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(CREATE_FIELDS),
            ),
            instructions.core.ArrayInstruction(
                reg=Register(RegisterName.R, 1),
                address=Address(1),
            ),
            # tp arg
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(1),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 2),
                imm=Immediate(0),
            ),
            instructions.core.StoreInstruction(
                reg=Register(RegisterName.R, 1),
                entry=ArrayEntry(1, index=Register(RegisterName.R, 2)),
            ),
            # num pairs arg
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(1),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 2),
                imm=Immediate(1),
            ),
            instructions.core.StoreInstruction(
                reg=Register(RegisterName.R, 1),
                entry=ArrayEntry(1, index=Register(RegisterName.R, 2)),
            ),
            # create cmd
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(1),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 2),
                imm=Immediate(0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 3),
                imm=Immediate(1),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 4),
                imm=Immediate(0),
            ),
            instructions.core.CreateEPRInstruction(
                reg0=Register(RegisterName.R, 1),
                reg1=Register(RegisterName.R, 2),
                reg2=Register(RegisterName.C, 0),
                reg3=Register(RegisterName.R, 3),
                reg4=Register(RegisterName.R, 4),
            ),
            # wait cmd
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 2),
                imm=Immediate(OK_FIELDS),
            ),
            instructions.core.WaitAllInstruction(
                slice=ArraySlice(
                    address=Address(0),
                    start=Register(RegisterName.R, 1),
                    stop=Register(RegisterName.R, 2),
                ),
            ),
            # if statement
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(2),
            ),
            instructions.core.LoadInstruction(
                reg=Register(RegisterName.R, 0),
                entry=ArrayEntry(
                    address=Address(0),
                    index=Register(RegisterName.R, 1),
                ),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(0),
            ),
            instructions.core.BneInstruction(
                reg0=Register(RegisterName.R, 0),
                reg1=Register(RegisterName.R, 1),
                imm=Immediate(28),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(2),
            ),
            instructions.core.LoadInstruction(
                reg=Register(RegisterName.R, 0),
                entry=ArrayEntry(
                    address=Address(0),
                    index=Register(RegisterName.R, 1),
                ),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(1),
            ),
            instructions.core.AddInstruction(
                reg0=Register(RegisterName.R, 0),
                reg1=Register(RegisterName.R, 0),
                reg2=Register(RegisterName.R, 1),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 1),
                imm=Immediate(2),
            ),
            instructions.core.StoreInstruction(
                reg=Register(RegisterName.R, 0),
                entry=ArrayEntry(
                    address=Address(0),
                    index=Register(RegisterName.R, 1),
                ),
            ),
            # return cmds
            instructions.core.RetArrInstruction(
                address=Address(0),
            ),
            instructions.core.RetArrInstruction(
                address=Address(1),
            ),
        ],
    )
    for i, instr in enumerate(subroutine.instructions):
        print(repr(instr))
        expected_instr = expected.instructions[i]
        print(repr(expected_instr))
        print()
        assert instr == expected_instr
    print(subroutine)
    print(expected)


def test_epr_r_create():

    set_log_level(logging.DEBUG)

    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        outcomes = epr_socket.create(tp=EPRType.R)
        m = outcomes[0].raw_measurement_outcome
        with m.if_eq(0):
            m.add(1)

    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
    assert len(alice.storage) == 5
    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)

    expected_text = """
# NETQASM 0.0
# APPID 0
set R1 10
array R1 @0
set R1 20
array R1 @1
set R1 2
set R2 0
store R1 @1[R2]
set R1 1
set R2 1
store R1 @1[R2]
set R1 1
set R2 0
set R3 1
set R4 0
create_epr R1 R2 C0 R3 R4
set R1 0
set R2 10
wait_all @0[R1:R2]
set R1 2
load R0 @0[R1]
set R1 0
bne R0 R1 28
set R1 2
load R0 @0[R1]
set R1 1
add R0 R0 R1
set R1 2
store R0 @0[R1]
ret_arr @0
ret_arr @1
    """

    expected = parse_text_subroutine(expected_text)

    for i, instr in enumerate(subroutine.instructions):
        print(repr(instr))
        expected_instr = expected.instructions[i]
        print(repr(expected_instr))
        print()
        assert instr == expected_instr
    print(f"subroutine: {subroutine}")
    print(f"expected: {expected}")


def test_epr_r_receive():

    set_log_level(logging.DEBUG)

    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        _ = epr_socket.recv(tp=EPRType.R)[0]

    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
    assert len(alice.storage) == 5
    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)
    print(subroutine)

    expected_text = """
# NETQASM 0.0
# APPID 0
set R5 10
array R5 @0
set R5 1
array R5 @1
set R5 0
set R6 0
store R5 @1[R6]
set R5 1
set R6 0
set R7 1
set R8 0
recv_epr R5 R6 R7 R8
set R5 0
set R6 10
wait_all @0[R5:R6]
set R2 0
set R5 1
beq R2 R5 42
load R0 @1[R2]
set R3 9
set R4 0
beq R4 R2 27
set R5 10
add R3 R3 R5
set R5 1
add R4 R4 R5
jmp 21
load R1 @0[R3]
set R0 0
set R5 3
bne R1 R5 32
rot_z R0 16 4
set R5 1
bne R1 R5 35
rot_x R0 16 4
set R5 2
bne R1 R5 39
rot_x R0 16 4
rot_z R0 16 4
set R5 1
add R2 R2 R5
jmp 16
ret_arr @0
ret_arr @1
    """

    expected = parse_text_subroutine(expected_text)

    for i, instr in enumerate(subroutine.instructions):
        print(f"actual command: {repr(instr)}")
        expected_instr = expected.instructions[i]
        print(f"expected command: {repr(expected_instr)}")
        print()
        assert instr == expected_instr
    print(f"actual: {subroutine}")
    print(f"expected: {expected}")


def test_epr_max_time():

    set_log_level(logging.DEBUG)

    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        q1 = epr_socket.create(time_unit=TimeUnit.MILLI_SECONDS, max_time=25)[0]
        q1.H()

    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
    assert len(alice.storage) == 5
    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine)
    print(subroutine)

    expected_text = """
# NETQASM 0.0
# APPID 0
set R0 10
array R0 @0
set R0 1
array R0 @1
set R0 0
set R1 0
store R0 @1[R1]
set R0 20
array R0 @2
set R0 0
set R1 0
store R0 @2[R1]
set R0 1
set R1 1
store R0 @2[R1]
set R0 1
set R1 5
store R0 @2[R1]
set R0 25
set R1 6
store R0 @2[R1]
set R0 1
set R1 0
set R2 1
set R3 2
set R4 0
create_epr R0 R1 R2 R3 R4
set R0 0
set R1 10
wait_all @0[R0:R1]
set Q0 0
h Q0
ret_arr @0
ret_arr @1
ret_arr @2
    """

    expected = parse_text_subroutine(expected_text)

    for i, instr in enumerate(subroutine.instructions):
        print(repr(instr))
        expected_instr = expected.instructions[i]
        print(repr(expected_instr))
        assert instr == expected_instr
    print(subroutine)
    print(expected)


if __name__ == "__main__":
    test_simple()
    test_rotations()
    test_epr_k_create()
    test_epr_k_recv()
    test_two_epr_k_create()
    test_two_epr_k_recv()
    test_epr_m()
    test_epr_r_create()
    test_epr_r_receive()
    test_epr_max_time()



from netqasm.runtime.application import Application, ApplicationInstance, Program
from netqasm.runtime.interface.config import default_network_config
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.external import simulate_application


def client():
    with DebugConnection("client") as conn:
        print("this is the client")
        builder = conn.builder

        q = builder.new_qubit()
        r = builder.new_register()
        a = builder.new_array()
        builder.measure(q)

        print(builder.print_instrs())


def run():
    prog_client = Program(party="client", entry=client, args=["app_config"], results=[])
    network_cfg = default_network_config(["delft"])

    app = Application(programs=[prog_client], metadata=None)
    app_instance = ApplicationInstance(
        app=app,
        program_inputs={
            "client": {},
        },
        network=None,
        party_alloc={
            "client": "delft",
        },
        logging_cfg=None,
    )

    simulate_application(
        app_instance=app_instance, network_cfg=network_cfg, use_app_config=False
    )


if __name__ == "__main__":
    run()



import numpy as np
import pytest

from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.runtime.settings import Simulator, get_simulator
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()


def run_alice():
    epr_socket = EPRSocket("Bob")
    with NetQASMConnection("Alice", epr_sockets=[epr_socket]):
        # Create entanglement
        epr_socket.create()[0]


def run_bob():
    epr_socket = EPRSocket("Alice")
    with NetQASMConnection("Bob", epr_sockets=[epr_socket]):
        epr_socket.recv()


def post_function(backend):
    alice_state = backend.nodes["Alice"].qmemory._get_qubits(0)[0].qstate
    bob_state = backend.nodes["Bob"].qmemory._get_qubits(0)[0].qstate
    assert alice_state is bob_state
    expected_state = np.array(
        [[0.5, 0, 0, 0.5], [0, 0, 0, 0], [0, 0, 0, 0], [0.5, 0, 0, 0.5]]
    )

    logger.info(f"state = {alice_state.qrepr.reduced_dm()}")
    assert np.all(np.isclose(expected_state, alice_state.qrepr.reduced_dm()))


@pytest.mark.skipif(
    get_simulator() == Simulator.SIMULAQRON,
    reason="SimulaQron does not yet support a post_function",
)
def test_create_epr():
    # app_instance = default_app_instance([
    #     ("Alice", run_alice),
    #     ("Bob", run_bob),
    # ])
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
            ("Bob", run_bob),
        ]
    )
    simulate_application(
        app_instance,
        use_app_config=False,
        post_function=post_function,
        enable_logging=False,
    )


if __name__ == "__main__":
    test_create_epr()



from netqasm.sdk.epr_socket import EPRSocket
from netqasm.sdk.external import NetQASMConnection


def main(app_config=None):
    log_config = app_config.log_config

    epr_socket = EPRSocket("server")

    client = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=log_config,
        epr_sockets=[epr_socket],
        # return_arrays=False,
    )

    with client:
        num_pairs = 10
        outcomes = client.new_array(num_pairs)

        def post_create(conn, q, pair):
            array_entry = outcomes.get_future_index(pair)
            # store measurement outcome in array
            q.measure(array_entry)

        # Create EPR pair
        epr_socket.create_keep(
            number=num_pairs,
            sequential=True,
            post_routine=post_create,
        )

    bits = "".join([str(outcome) for outcome in outcomes])
    print(bits)


if __name__ == "__main__":
    main()



class Symbols:
    COMMENT_START = "//"
    BRANCH_END = ":"
    MACRO_START = "$"
    ADDRESS_START = "@"
    ARGS_BRACKETS = "()"
    ARGS_DELIM = ","
    INDEX_BRACKETS = "[]"
    SLICE_DELIM = ":"
    TEMPLATE_BRACKETS = "{}"

    PREAMBLE_START = "#"
    PREAMBLE_NETQASM = "NETQASM"
    PREAMBLE_APPID = "APPID"
    PREAMBLE_DEFINE = "DEFINE"
    PREAMBLE_DEFINE_BRACKETS = r"{}"



from netqasm.sdk.qubit import Qubit


def t_inverse(q: Qubit) -> None:
    """Performs an inverse of the T gate by applying the T gate 7 times."""
    for _ in range(7):
        q.T()


def toffoli_gate(control1: Qubit, control2: Qubit, target: Qubit) -> None:
    """Performs a Toffoli gate with `control1` and `control2` as control qubits
    and `target` as target, using CNOTS, Ts and Hadamard gates.

    See https://en.wikipedia.org/wiki/Toffoli_gate
    """
    target.H()
    control2.cnot(target)
    t_inverse(target)
    control1.cnot(target)
    target.T()
    control2.cnot(target)
    t_inverse(target)
    control1.cnot(target)
    control2.T()
    target.T()
    target.H()
    control1.cnot(control2)
    control1.T()
    t_inverse(control2)
    control1.cnot(control2)



import numpy as np

from netqasm.examples.lib.bqc import recv_meas_outcome, send_meas_cmd, teleport_state
from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket

logger = get_netqasm_logger()


def get_phi_for_oracle(b0, b1):
    """Compute the angles `phi1` and `phi2` needed to simulate
    an oracle that only tags the input (b0, b1).
    """
    phi1 = np.pi / 2 - b0 * np.pi
    phi2 = (1 - (b0 ^ b1)) * np.pi
    return phi1, phi2


def main(app_config=None, b0=0, b1=0, r1=0, r2=0, theta1=0.0, theta2=0.0):

    socket = Socket("client", "server", log_config=app_config.log_config)
    epr_socket = EPRSocket("server")

    client = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )

    num_qubits = 4
    phi1, phi2 = get_phi_for_oracle(b0, b1)

    # Set theta0 and theta3 to 0.
    theta = [0 for _ in range(num_qubits)]
    theta[1] = theta1
    theta[2] = theta2

    with client:
        # Teleport states q0 to q3 to the Server.
        # The resulting state q[i] might have a phase `pi`,
        # depending on outcome m[i].
        m = [None] * num_qubits
        for i in range(num_qubits):
            m[i] = teleport_state(epr_socket, theta[i])
        client.flush()

        # Convert outcomes to integers to use them in calculations below.
        m = [int(m[i]) for i in range(num_qubits)]

        # Let the Server measure q1. We want to measure with angle phi1,
        # but send delta1 instead, which compensates for m1, r1 and theta1.
        delta1 = phi1 - theta[1] + r1 * np.pi - m[1] * np.pi
        send_meas_cmd(socket, delta1)
        s1 = recv_meas_outcome(socket)

        # Let the Server measure q2. We want to measure with angle phi2,
        # but send delta2 instead, which compensates for m1, s1, r1, and theta2.
        delta2 = phi2 - theta[2] + (s1 ^ r1) * np.pi + r2 * np.pi - m[2] * np.pi
        send_meas_cmd(socket, delta2)
        s2 = recv_meas_outcome(socket)

        # At this point, and before the Server measures both output qubits (q0 and q3)
        # in the Y basis, there are still some Pauli byproducts.
        # For q0, these byproducts are Z^m0 X^s1 X^r1.
        # For q3, these byproducts are Z^m3 X^s2 X^r2.
        # However, since these all commute with Y, we will simply let
        # the Server measure Y anyway, and apply bit-flips afterwards.
        result0 = recv_meas_outcome(socket)
        result1 = recv_meas_outcome(socket)

        # Flip bits according to Pauli byproducts (^ = xor).
        if (s1 ^ r1 ^ m[0]) == 1:
            result0 = 1 - result0
        if (s2 ^ r2 ^ m[3]) == 1:
            result1 = 1 - result1

        return {
            "result0": result0,
            "result1": result1,
            "phi1": phi1,
            "phi2": phi2,
            "delta1": delta1,
            "delta2": delta2,
            "s1": s1,
            "s2": s2,
            "m": m,
            "b0": b0,
            "b1": b1,
            "r1": r1,
            "r2": r2,
            "theta1": theta[1],
            "theta2": theta[2],
        }



import numpy as np

from netqasm.logging.output import get_new_app_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket


def measure_basis_0(bob, q):
    q.rot_Y(angle=-np.pi / 4)
    return q.measure()


def measure_basis_1(bob, q):
    q.rot_Y(angle=np.pi / 4)
    return q.measure()


def main(app_config=None, y=0):
    if not (y == 0 or y == 1):
        raise ValueError(f"y should be 0 or 1, not {y}")

    app_logger = get_new_app_logger(
        app_name=app_config.app_name, log_config=app_config.log_config
    )
    app_logger.log(f"Bob received input bit y = {y}")

    socket = Socket("bob", "repeater", log_config=app_config.log_config)
    epr_socket = EPRSocket("repeater")

    bob = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )

    with bob:
        # Wait for entanglement with Alice through repeater
        epr = epr_socket.recv_keep()[0]
        bob.flush()

        # Receive teleportation corrections
        msg = socket.recv()
        app_logger.log(f"Bob got teleportation corrections: {msg}")
        m1, m2 = eval(msg)
        if m2 == 1:
            epr.X()
        if m1 == 1:
            epr.Z()

        # CHSH strategy: measure in one of 2 bases depending on y.
        if y == 0:
            b = measure_basis_0(bob, epr)
        else:
            b = measure_basis_1(bob, epr)

    app_logger.log(f"Bob outputs b = {b}")
    return {"b": int(b)}


if __name__ == "__main__":
    main()



import random

from netqasm.logging.glob import set_log_level
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.qubit import Qubit


def main(no_output=False):
    with DebugConnection("Alice") as alice:
        num = 10

        outcomes = alice.new_array(num)
        rand_nums = alice.new_array(
            init_values=[random.randint(0, 1) for _ in range(num)]
        )

        with rand_nums.enumerate() as (i, r):
            q = Qubit(alice)
            with r.if_eq(1):
                q.X()
            q.measure(future=outcomes.get_future_index(i))

    if no_output:
        print(f"binary:\n{alice.storage[0]}")


if __name__ == "__main__":
    set_log_level("INFO")
    main()



from netqasm.runtime.settings import Simulator, get_simulator
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket, get_qubit_state
from netqasm.sdk.toolbox.sim_states import get_fidelity, qubit_from, to_dm


def main(app_config=None):
    log_config = app_config.log_config

    # Create a socket to recv classical information
    socket = Socket("receiver", "sender", log_config=log_config)

    # Create a EPR socket for entanglement generation
    epr_socket = EPRSocket("sender")

    # Initialize the connection
    receiver = NetQASMConnection(
        app_name=app_config.app_name, log_config=log_config, epr_sockets=[epr_socket]
    )
    with receiver:
        epr = epr_socket.recv_keep()[0]
        receiver.flush()

        # Get the corrections
        m1, m2 = socket.recv_structured().payload
        print(f"`receiver` got corrections: {m1}, {m2}")
        if m2 == 1:
            print("`receiver` will perform X correction")
            epr.X()
        if m1 == 1:
            print("`receiver` will perform Z correction")
            epr.Z()

        receiver.flush()

        if get_simulator() == Simulator.NETSQUID:
            # Get the qubit state
            # NOTE only possible in simulation, not part of actual application
            dm = get_qubit_state(epr)
            print(f"`receiver` recieved the teleported state {dm}")

            # Reconstruct the original qubit to compare with the received one.
            # NOTE only to check simulation results, normally the Sender does not
            # need to send the phi and theta values!
            msg = socket.recv_silent()  # don't log this
            print(f"received silent message: {msg}")
            phi, theta = eval(msg)

            original = qubit_from(phi, theta)
            original_dm = to_dm(original)
            fidelity = get_fidelity(original, dm)

            return {
                "original_state": original_dm.tolist(),
                "correction1": "Z" if m1 == 1 else "None",
                "correction2": "X" if m2 == 1 else "None",
                "received_state": dm.tolist(),
                "fidelity": fidelity,
            }
        else:
            return {
                "correction1": "Z" if m1 == 1 else "None",
                "correction2": "X" if m2 == 1 else "None",
            }


if __name__ == "__main__":
    main()



from yaml import dump, load

try:
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Dumper  # type: ignore
    from yaml import Loader  # type: ignore


def load_yaml(file_path):
    with open(file_path, "r") as f:
        data = load(f, Loader=Loader)
    return data


def dump_yaml(data, file_path):
    with open(file_path, "w") as f:
        dump(data, f, Dumper=Dumper)



import os

from click.testing import CliRunner

import netqasm
from netqasm.runtime.cli import cli
from netqasm.runtime.env import EXAMPLE_APPS_DIR
from netqasm.util.yaml import load_yaml

TEMPLATE_EXAMPLE_NAME = "teleport"
TEMPLATE_EXAMPLE_DIR = os.path.join(EXAMPLE_APPS_DIR, TEMPLATE_EXAMPLE_NAME)

IGNORED_FILES = [
    "__init__.py",
    "__pycache__",
    "log",
    "cysignals_crash_logs",
]


def test_version():
    runner = CliRunner()
    results = runner.invoke(cli, "version")
    print(results.output)
    assert results.exit_code == 0
    assert results.output.strip() == netqasm.__version__


def test_new():
    runner = CliRunner()
    with runner.isolated_filesystem():
        path = "test"
        results = runner.invoke(cli, ["new", path])
        print(results.output)
        assert results.exit_code == 0
        assert results.output.startswith("Creating application")
        assert TEMPLATE_EXAMPLE_NAME in results.output
        expected_files = [
            f for f in os.listdir(TEMPLATE_EXAMPLE_DIR) if f not in IGNORED_FILES
        ]
        assert sorted(os.listdir(path)) == sorted(expected_files)


def test_quiet():
    runner = CliRunner()
    with runner.isolated_filesystem():
        path = "test"
        results = runner.invoke(cli, ["new", "-q", path])
        print(results.output)
        assert results.exit_code == 0
        assert results.output == ""


def test_new_existing():
    runner = CliRunner()
    with runner.isolated_filesystem():
        path = "test"
        results = runner.invoke(cli, ["new", path])
        assert results.exit_code == 0
        results = runner.invoke(cli, ["new", path])
        print(results.output)
        assert results.exit_code != 0
        assert "already exists" in results.output


def test_new_template():
    runner = CliRunner()
    with runner.isolated_filesystem():
        path = "test"
        template = "anonymous_transmission"
        template_example_dir = os.path.join(EXAMPLE_APPS_DIR, template)
        results = runner.invoke(cli, ["new", path, f"--template={template}"])
        print(results.output)
        assert results.exit_code == 0
        assert results.output.startswith("Creating application")
        assert template in results.output
        expected_files = [
            f for f in os.listdir(template_example_dir) if f not in IGNORED_FILES
        ]
        assert sorted(os.listdir(path)) == sorted(expected_files)


# Test disabled for now. TODO: fix this
def disabled_test_init():
    runner = CliRunner()
    with runner.isolated_filesystem():
        # Create template
        path = "test"
        results = runner.invoke(cli, ["new", path])
        results.exit_code == 0
        files_start = os.listdir(path)
        # Remove all config files
        for entry in os.listdir(path):
            if not (entry.startswith("app_") and entry.endswith(".py")):
                os.remove(os.path.join(path, entry))
        assert len(os.listdir(path)) == 2

        # Initialize again
        results = runner.invoke(cli, ["init", f"--path={path}"])
        print(results.output)
        assert results.exit_code == 0
        # Check that it's the same number of files again
        assert sorted(os.listdir(path)) == sorted(files_start)

        # Check content based on teleport example
        expected_nodes = ["sender", "receiver"]
        # network
        network = load_yaml(os.path.join(path, "network.yaml"))
        assert "nodes" in network
        nodes = [node["name"] for node in network["nodes"]]
        assert sorted(expected_nodes) == sorted(nodes)
        assert "links" in network
        assert len(network["links"]) == 1
        # roles
        roles = load_yaml(os.path.join(path, "roles.yaml"))
        assert sorted(expected_nodes) == sorted(roles.keys())
        # input
        sender_input = load_yaml(os.path.join(path, "sender.yaml"))
        assert "phi" in sender_input
        assert "theta" in sender_input
        receiver_input = load_yaml(os.path.join(path, "receiver.yaml"))
        assert len(receiver_input) == 0


def test_init_not_app_dir():
    runner = CliRunner()
    with runner.isolated_filesystem():
        results = runner.invoke(cli, ["init"])
        print(results.output)
        print(results.exc_info)
        assert isinstance(results.exception, ValueError)
        assert "does not seem to be" in str(results.exception)
        assert results.exit_code != 0


def test_init_no_overwrite():
    runner = CliRunner()
    with runner.isolated_filesystem():
        path = "test"
        results = runner.invoke(cli, ["new", path])
        # Write test to all files
        for entry in os.listdir(path):
            with open(os.path.join(path, entry), "w") as f:
                f.write("test")

        results = runner.invoke(cli, ["init", f"--path={path}"])
        assert results.exit_code == 0
        assert results.output.startswith("No files needed to be added")

        # Check that files remained
        for entry in os.listdir(path):
            with open(os.path.join(path, entry), "r") as f:
                assert f.read() == "test"


# class TestQNE(unittest.TestCase):
#     """Test interaction using a mocked apirouter."""

#     username = os.environ.get("QNE_TEST_USER", "test_user")
#     password = os.environ.get("QNE_TEST_PWD", "test_pass")
#     path = ".qne"
#     jwt_token = "abcdefgh"
#     api_token = 12345678
#     mock_get_patcher = NotImplemented  # type: unittest.mock._patch
#     mock_post_patcher = NotImplemented  # type: unittest.mock._patch
#     mock_path_patcher = NotImplemented  # type: unittest.mock._patch
#     mock_get = NotImplemented  # type: unittest.mock.Mock
#     mock_post = NotImplemented  # type: unittest.mock.Mock

#     @classmethod
#     def setUpClass(cls) -> None:
#         cls.mock_get_patcher = patch("netqasm.runtime.cli.requests.get")
#         cls.mock_post_patcher = patch("netqasm.runtime.cli.requests.post")
#         cls.mock_path_patcher = patch("netqasm.runtime.cli.QNE_FOLDER_PATH", cls.path)
#         cls.mock_path_patcher.start()
#         cls.mock_get = cls.mock_get_patcher.start()
#         cls.mock_post = cls.mock_post_patcher.start()

#         def get_imitator(url, headers=None):
#             response = requests.Response()
#             response.status_code = 504
#             return response

#         cls.mock_get.side_effect = get_imitator

#         def post_imitator(url, data, headers=None):
#             response = requests.Response()
#             if re.match(r".*/jwt/", url):
#                 if (
#                     data["username"][0] == cls.username
#                     and data["password"][0] == cls.password
#                 ):
#                     response.status_code = 200
#                     response._content = json.dumps({"access": cls.jwt_token}).encode(
#                         "utf-8"
#                     )
#                 else:
#                     response.status_code = 401
#             elif re.match(r".*/auth/token/destroy/", url):
#                 response.status_code = 204
#             elif re.match(r".*/auth/token/create/", url):
#                 if headers["Authorization"] != f"JWT {cls.jwt_token}":
#                     response.status_code = 401
#                 else:
#                     # Username is actually not used, you can't create tokens for others.
#                     response.status_code = 200
#                     cls.api_token += 1
#                     response._content = json.dumps(
#                         {"access": f"{cls.api_token}"}
#                     ).encode("utf-8")
#             else:
#                 response.status_code = 504
#             return response

#         cls.mock_post.side_effect = post_imitator

#     @classmethod
#     def tearDownClass(cls) -> None:
#         cls.mock_get_patcher.stop()
#         cls.mock_post_patcher.stop()
#         cls.mock_path_patcher.stop()

#     def test_qne_login(self):
#         """Test logging into the QNE produces a token."""
#         runner = CliRunner()
#         api_file = f"{self.path}/api_token"
#         with runner.isolated_filesystem():
#             assert not os.path.exists(api_file)
#             for args in [
#                 (self.username, ""),
#                 ("", self.password),
#                 (True, self.password),
#                 (self.username, True),
#                 (self.username, self.password, True),
#             ]:
#                 try:
#                     _login(*args)
#                 except (ValueError, TypeError):
#                     assert not os.path.exists(api_file)
#             # Test authentication doesn't crash if invalid credentials are supplied.
#             results = runner.invoke(
#                 cli,
#                 ["--verbose", "qne", "login"],
#                 input=f"{self.username}\nwrongpass\n",
#             )
#             assert results.exit_code == 0
#             assert not os.path.exists(api_file)
#             # Test we can authenticate successfully
#             results = runner.invoke(
#                 cli,
#                 ["--verbose", "qne", "login"],
#                 input=f"{self.username}\n{self.password}\n",
#             )
#             assert results.exit_code == 0
#             assert os.path.exists(api_file)
#             with open(api_file) as f:
#                 api_token = json.load(f)
#             assert self.username == list(api_token.values())[0][0]
#             # Test we get new tokens
#             results = runner.invoke(
#                 cli,
#                 ["--verbose", "qne", "login"],
#                 input=f"{self.username}\n{self.password}\n",
#             )
#             assert results.exit_code == 0
#             with open(api_file) as f:
#                 new_api_token = json.load(f)
#             assert self.username == list(new_api_token.values())[0][0]
#             assert api_token != new_api_token
#             # Test we can run in verbose mode
#             results = runner.invoke(
#                 cli,
#                 ["--verbose", "qne", "login"],
#                 input=f"{self.username}\n{self.password}\n",
#             )
#             assert results.exit_code == 0

#     def test_qne_logout(self):
#         """Test logging out."""
#         runner = CliRunner()
#         with runner.isolated_filesystem():
#             # Create an API token
#             results = runner.invoke(
#                 cli, ["qne", "login"], input=f"{self.username}\n{self.password}\n"
#             )
#             assert results.exit_code == 0
#             assert os.path.exists(f"{self.path}/api_token")
#             # Test API token is destroyed when logged out.
#             results = runner.invoke(cli, ["qne", "logout"])
#             assert results.exit_code == 0
#             assert not os.path.exists(f"{self.path}/api_token")
#             # Test logging out when already logged out doesn't crash.
#             results = runner.invoke(cli, ["qne", "logout"])
#             assert results.exit_code == 0
#             assert not os.path.exists(f"{self.path}/api_token")



from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, get_qubit_state
from netqasm.sdk.toolbox import set_qubit_state

from .conf import apps
from .sub_protocols import quantum_anonymous_tranmission, setup_sockets

logger = get_netqasm_logger()


def anonymous_transmission(
    app_name,
    app_config=None,
    sender=False,
    receiver=False,
    phi=None,
    theta=None,
):

    # Setup sockets, epr_sockets and a broadcast_channel needed for the protocol
    sockets = setup_sockets(
        app_name=app_config.app_name,
        apps=apps,
        log_config=app_config.log_config,
    )

    # Initialize the connection to the backend
    conn = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=sockets.epr_sockets,
    )
    with conn:
        if sender:
            # Prepare the qubit to teleport
            qubit = Qubit(conn)
            set_qubit_state(qubit=qubit, phi=phi, theta=theta)
        else:
            qubit = None
        q = quantum_anonymous_tranmission(
            conn=conn,
            sockets=sockets,
            num_nodes=len(apps),
            sender=sender,
            receiver=receiver,
            qubit=qubit,
        )
        if receiver:
            dm = get_qubit_state(q)
            output = {"qubit_state": dm.tolist()}
        else:
            output = None
    return output



"""Transpilation of subroutines from one flavour to another.

This module contains the SubroutineTranspiler interface which is a base class for
transpilers that can convert a NetQASM subroutine into a subroutine with a specific
flavour.
"""

import abc
from typing import Dict, List, Optional, Set, Tuple, Union

from netqasm.lang.instr import DebugInstruction, NetQASMInstruction, core, nv, vanilla
from netqasm.lang.operand import Immediate, Register, RegisterName
from netqasm.lang.subroutine import Subroutine
from netqasm.runtime.settings import get_is_using_hardware
from netqasm.util.log import HostLine


class SubroutineTranspiler(abc.ABC):
    def __init__(self, subroutine: Subroutine, debug: bool = False):
        pass

    @abc.abstractmethod
    def transpile(self) -> Subroutine:
        """Transpile the subroutine into one with a specific flavour."""
        pass


class NVSubroutineTranspiler(SubroutineTranspiler):
    """A transpiler that converts a subroutine with the vanilla flavour to a subroutine
    with the NV flavour.
    """

    def __init__(self, subroutine: Subroutine, debug=False):
        self._subroutine: Subroutine = subroutine
        self._used_registers: Set[Register] = set()
        self._register_values: Dict[Register, Immediate] = dict()
        self._debug: bool = debug

    def get_reg_value(self, reg: Register) -> Immediate:
        """Get the value of a register at this moment"""
        return self._register_values[reg]

    def get_unused_register(self) -> Register:
        """
        Naive approach: try to use Q0 if possible, otherwise Q1, etc.
        """
        for i in range(16):
            reg = Register(RegisterName.Q, i)
            if reg not in self._used_registers:
                return reg
        raise RuntimeError("Could not find free register")

    def swap(
        self,
        lineno: Optional[HostLine],
        electron: Register,
        carbon: Register,
    ) -> List[NetQASMInstruction]:
        """
        Swap the states of the electron and a carbon.
        See https://gitlab.tudelft.nl/qinc-wehner/netqasm/netqasm-docs/-/blob/master/nv-gates-docs.md
        for the circuit.
        """
        gates: List[NetQASMInstruction] = []

        if self._debug:
            gates += [DebugInstruction(text="begin SWAP")]

        gates += [
            nv.ControlledRotXInstruction(
                lineno=lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(8),
                imm1=Immediate(4),
            ),
            nv.RotXInstruction(
                lineno=lineno, reg=electron, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.RotYInstruction(
                lineno=lineno, reg=electron, imm0=Immediate(16), imm1=Immediate(4)
            ),
            nv.RotZInstruction(
                lineno=lineno, reg=carbon, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.ControlledRotXInstruction(
                lineno=lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(8),
                imm1=Immediate(4),
            ),
            nv.RotXInstruction(
                lineno=lineno, reg=electron, imm0=Immediate(8), imm1=Immediate(4)
            ),
            nv.RotYInstruction(
                lineno=lineno, reg=electron, imm0=Immediate(8), imm1=Immediate(4)
            ),
            nv.RotXInstruction(
                lineno=lineno, reg=carbon, imm0=Immediate(8), imm1=Immediate(4)
            ),
            nv.RotZInstruction(
                lineno=lineno, reg=carbon, imm0=Immediate(8), imm1=Immediate(4)
            ),
            nv.ControlledRotXInstruction(
                lineno=lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(8),
                imm1=Immediate(4),
            ),
            nv.RotYInstruction(
                lineno=lineno, reg=electron, imm0=Immediate(16), imm1=Immediate(4)
            ),
            nv.RotZInstruction(
                lineno=lineno, reg=carbon, imm0=Immediate(16), imm1=Immediate(4)
            ),
        ]

        if self._debug:
            gates += [DebugInstruction(text="end SWAP")]

        return gates

    def transpile(self) -> Subroutine:
        """
        Very simple transpiling pass: iterate over all instructions once and rewrite them in-line.
        While iterating, keep track of which registers are in use and what their values are.
        """
        new_commands: List[NetQASMInstruction] = []

        index_changes = {}  # map index in commands to index in new_commands

        for i, instr in enumerate(self._subroutine.instructions):
            # check which registers are being written to
            affected_regs = instr.writes_to()

            for reg in affected_regs:
                if reg.name != RegisterName.Q:
                    continue  # for now we are only interested in Q-register values

                if isinstance(instr, core.SetInstruction):
                    # OK, value is a known Immediate. Update register value:
                    self._register_values[reg] = instr.imm
                else:
                    pass
                    # don't allow writing to a Q-register by any other instruction type
                    # TODO
                    # raise RuntimeError(
                    #     f"Cannot transpile: the instruction {instr} writes to"
                    #     " a Q-register but the value cannot be determined"
                    #     " at transpile time.")

            for op in instr.operands:
                # update used registers
                if isinstance(op, Register):
                    self._used_registers.update([op])

            index_changes[i] = len(new_commands)

            if isinstance(instr, core.SingleQubitInstruction) or isinstance(
                instr, core.RotationInstruction
            ):
                new_commands += self._handle_single_qubit_gate(instr)
            elif isinstance(instr, core.TwoQubitInstruction):
                new_commands += self._handle_two_qubit_gate(instr)
            else:
                new_commands += [instr]

        add_no_op_at_end = False

        for instr in new_commands:
            if (
                isinstance(instr, core.BranchUnaryInstruction)
                or isinstance(instr, core.BranchBinaryInstruction)
                or isinstance(instr, core.JmpInstruction)
            ):
                original_line = instr.line.value
                if original_line == len(self._subroutine.instructions):
                    # There was a label in the original subroutine at the very end.
                    # Since this label is now removed, we should put a "no-op"
                    # instruction there so there is something to jump to.
                    add_no_op_at_end = True
                    instr.line = Immediate(len(new_commands))
                else:
                    instr.line = Immediate(index_changes[instr.line.value])

        if add_no_op_at_end:
            new_commands += [
                core.SetInstruction(
                    lineno=None, reg=Register(RegisterName.C, 15), imm=Immediate(1337)
                )
            ]

        self._subroutine.instructions = new_commands
        return self._subroutine

    def _move_electron_carbon(
        self, instr: vanilla.MovInstruction
    ) -> List[NetQASMInstruction]:
        """
        See https://gitlab.tudelft.nl/qinc-wehner/netqasm/netqasm-docs/-/blob/master/nv-gates-docs.md
        for the circuit.
        """
        electron = instr.reg0
        carbon = instr.reg1
        return [
            nv.RotYInstruction(
                lineno=instr.lineno, reg=electron, imm0=Immediate(8), imm1=Immediate(4)
            ),
            nv.ControlledRotYInstruction(
                lineno=instr.lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(24),
                imm1=Immediate(4),
            ),
            nv.RotXInstruction(
                lineno=instr.lineno, reg=electron, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.ControlledRotXInstruction(
                lineno=instr.lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(8),
                imm1=Immediate(4),
            ),
        ]

    def _move_carbon_electron(
        self, instr: vanilla.MovInstruction
    ) -> List[NetQASMInstruction]:
        """
        See https://gitlab.tudelft.nl/qinc-wehner/netqasm/netqasm-docs/-/blob/master/nv-gates-docs.md
        for the circuit.
        """
        electron = instr.reg1
        carbon = instr.reg0
        return [
            nv.RotYInstruction(
                lineno=instr.lineno, reg=electron, imm0=Immediate(8), imm1=Immediate(4)
            ),
            nv.ControlledRotYInstruction(
                lineno=instr.lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(24),
                imm1=Immediate(4),
            ),
            nv.RotXInstruction(
                lineno=instr.lineno, reg=electron, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.ControlledRotXInstruction(
                lineno=instr.lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(8),
                imm1=Immediate(4),
            ),
            nv.RotYInstruction(
                lineno=instr.lineno, reg=electron, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.RotZInstruction(
                lineno=instr.lineno, reg=electron, imm0=Immediate(24), imm1=Immediate(4)
            ),
        ]

    def _handle_two_qubit_gate(
        self, instr: core.TwoQubitInstruction
    ) -> List[NetQASMInstruction]:
        try:
            qubit_id0 = self.get_reg_value(instr.reg0).value
            qubit_id1 = self.get_reg_value(instr.reg1).value
        except KeyError:
            # Register values are not known at transpile time.
            # We may assume that this is a MOV operation from the communication
            # qubit to a memory qubit. (This is the only time a gate uses
            # operands that are not known at transpile time.)
            assert isinstance(instr, vanilla.MovInstruction)
            return self._move_electron_carbon(instr)

        assert qubit_id0 != qubit_id1

        # It is assumed that there is only one electron, and that its virtual ID is 0.
        if isinstance(instr, vanilla.CnotInstruction):
            if qubit_id0 == 0:
                return self._map_cnot_electron_carbon(instr)
            elif qubit_id1 == 0:
                return self._map_cnot_carbon_electron(instr)
            else:
                return self._map_cnot_carbon_carbon(instr)
        elif isinstance(instr, vanilla.CphaseInstruction):
            if qubit_id0 == 0:
                return self._map_cphase_electron_carbon(instr)
            elif qubit_id1 == 0:
                swapped = vanilla.CphaseInstruction(
                    lineno=instr.lineno, reg0=instr.reg1, reg1=instr.reg0
                )
                return self._map_cphase_electron_carbon(swapped)
            else:
                return self._map_cphase_carbon_carbon(instr)
        elif isinstance(instr, vanilla.MovInstruction):
            if qubit_id0 == 0 and qubit_id1 != 0:
                return self._move_electron_carbon(instr)
            elif qubit_id0 != 0 and qubit_id1 == 0:
                return self._move_carbon_electron(instr)
            else:
                raise RuntimeError(f"Cannot move qubit {qubit_id0} to {qubit_id1}")
        else:
            raise ValueError(
                f"Don't know how to map instruction {instr} of type {type(instr)}"
            )

    def _map_cphase_electron_carbon(
        self,
        instr: vanilla.CphaseInstruction,
    ) -> List[NetQASMInstruction]:
        electron = instr.reg0
        carbon = instr.reg1
        """
        See https://gitlab.tudelft.nl/qinc-wehner/netqasm/netqasm-docs/-/blob/master/nv-gates-docs.md
        for the circuit.
        """

        return [
            nv.RotYInstruction(
                lineno=instr.lineno, reg=carbon, imm0=Immediate(8), imm1=Immediate(4)
            ),
            nv.ControlledRotXInstruction(
                lineno=instr.lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(8),
                imm1=Immediate(4),
            ),
            nv.RotZInstruction(
                lineno=instr.lineno, reg=electron, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.RotXInstruction(
                lineno=instr.lineno, reg=carbon, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.RotYInstruction(
                lineno=instr.lineno, reg=carbon, imm0=Immediate(24), imm1=Immediate(4)
            ),
        ]

    def _map_cphase_carbon_carbon(
        self, instr: vanilla.CphaseInstruction
    ) -> List[NetQASMInstruction]:
        """
        See https://gitlab.tudelft.nl/qinc-wehner/netqasm/netqasm-docs/-/blob/master/nv-gates-docs.md
        for the circuit.
        """
        electron = self.get_unused_register()
        carbon = instr.reg0
        set_electron = core.SetInstruction(
            lineno=instr.lineno, reg=electron, imm=Immediate(0)
        )
        instr.reg0 = electron

        result: List[NetQASMInstruction] = [set_electron]
        result += (
            self.swap(instr.lineno, electron, carbon)
            + self._map_cphase_electron_carbon(instr)
            + self.swap(instr.lineno, electron, carbon)
        )
        return result

    def _map_cnot_electron_carbon(
        self,
        instr: vanilla.CnotInstruction,
    ) -> List[NetQASMInstruction]:
        electron = instr.reg0
        carbon = instr.reg1
        """
        See https://gitlab.tudelft.nl/qinc-wehner/netqasm/netqasm-docs/-/blob/master/nv-gates-docs.md
        for the circuit.
        """

        return [
            nv.ControlledRotXInstruction(
                lineno=instr.lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(8),
                imm1=Immediate(4),
            ),
            nv.RotZInstruction(
                lineno=instr.lineno, reg=electron, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.RotXInstruction(
                lineno=instr.lineno, reg=carbon, imm0=Immediate(24), imm1=Immediate(4)
            ),
        ]

    def _map_cnot_carbon_electron(
        self,
        instr: vanilla.CnotInstruction,
    ) -> List[NetQASMInstruction]:
        """
        See https://gitlab.tudelft.nl/qinc-wehner/netqasm/netqasm-docs/-/blob/master/nv-gates-docs.md
        for the circuit.
        """
        electron = instr.reg1
        carbon = instr.reg0

        electron_hadamard = self._map_single_gate(
            instr=vanilla.GateHInstruction(lineno=instr.lineno, reg=electron)
        )

        gates = []
        gates += electron_hadamard
        gates += [
            nv.RotYInstruction(
                lineno=instr.lineno, reg=carbon, imm0=Immediate(8), imm1=Immediate(4)
            ),
            nv.ControlledRotXInstruction(
                lineno=instr.lineno,
                reg0=electron,
                reg1=carbon,
                imm0=Immediate(8),
                imm1=Immediate(4),
            ),
            nv.RotZInstruction(
                lineno=instr.lineno, reg=electron, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.RotXInstruction(
                lineno=instr.lineno, reg=carbon, imm0=Immediate(24), imm1=Immediate(4)
            ),
            nv.RotYInstruction(
                lineno=instr.lineno, reg=carbon, imm0=Immediate(24), imm1=Immediate(4)
            ),
        ]
        gates += electron_hadamard

        return gates

    def _map_cnot_carbon_carbon(
        self, instr: vanilla.CnotInstruction
    ) -> List[NetQASMInstruction]:
        """
        See https://gitlab.tudelft.nl/qinc-wehner/netqasm/netqasm-docs/-/blob/master/nv-gates-docs.md
        for the circuit.
        """
        electron = self.get_unused_register()
        carbon = instr.reg0
        set_electron = core.SetInstruction(
            lineno=instr.lineno, reg=electron, imm=Immediate(0)
        )
        instr.reg0 = electron

        result: List[NetQASMInstruction] = [set_electron]
        result += (
            self.swap(instr.lineno, electron, carbon)
            + self._map_cnot_electron_carbon(instr)
            + self.swap(instr.lineno, electron, carbon)
        )
        return result

    def _handle_single_qubit_gate(
        self,
        instr: Union[core.SingleQubitInstruction, core.RotationInstruction],
    ) -> List[NetQASMInstruction]:
        return self._map_single_gate(instr)

    def _map_single_gate(
        self,
        instr: Union[core.SingleQubitInstruction, core.RotationInstruction],
    ) -> List[NetQASMInstruction]:
        if isinstance(instr, vanilla.GateXInstruction):
            return [
                nv.RotXInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(16),
                    imm1=Immediate(4),
                )
            ]
        elif isinstance(instr, vanilla.GateYInstruction):
            return [
                nv.RotYInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(16),
                    imm1=Immediate(4),
                )
            ]
        elif isinstance(instr, vanilla.GateZInstruction):
            return [
                nv.RotXInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(24),
                    imm1=Immediate(4),
                ),
                nv.RotYInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(16),
                    imm1=Immediate(4),
                ),
                nv.RotXInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(8),
                    imm1=Immediate(4),
                ),
            ]
        elif isinstance(instr, vanilla.GateHInstruction):
            return [
                nv.RotYInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(8),
                    imm1=Immediate(4),
                ),
                nv.RotXInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(16),
                    imm1=Immediate(4),
                ),
            ]
        elif isinstance(instr, vanilla.GateKInstruction):
            return [
                nv.RotXInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(24),
                    imm1=Immediate(4),
                ),
                nv.RotYInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(16),
                    imm1=Immediate(4),
                ),
            ]
        elif isinstance(instr, vanilla.GateSInstruction):
            return [
                nv.RotXInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(24),
                    imm1=Immediate(4),
                ),
                nv.RotYInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(24),
                    imm1=Immediate(4),
                ),
                nv.RotXInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(8),
                    imm1=Immediate(4),
                ),
            ]
        elif isinstance(instr, vanilla.GateTInstruction):
            return [
                nv.RotXInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(24),
                    imm1=Immediate(4),
                ),
                nv.RotYInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(28),
                    imm1=Immediate(4),
                ),
                nv.RotXInstruction(
                    lineno=instr.lineno,
                    reg=instr.reg,
                    imm0=Immediate(8),
                    imm1=Immediate(4),
                ),
            ]
        elif isinstance(instr, vanilla.RotZInstruction):
            if get_is_using_hardware():
                imm0, imm1 = get_hardware_num_denom(instr)
            else:
                imm0, imm1 = instr.angle_num, instr.angle_denom
            return [
                nv.RotZInstruction(
                    lineno=instr.lineno, reg=instr.reg, imm0=imm0, imm1=imm1
                ),
            ]
        elif isinstance(instr, vanilla.RotXInstruction):
            if get_is_using_hardware():
                imm0, imm1 = get_hardware_num_denom(instr)
            else:
                imm0, imm1 = instr.angle_num, instr.angle_denom
            return [
                nv.RotXInstruction(
                    lineno=instr.lineno, reg=instr.reg, imm0=imm0, imm1=imm1
                ),
            ]
        elif isinstance(instr, vanilla.RotYInstruction):
            if get_is_using_hardware():
                imm0, imm1 = get_hardware_num_denom(instr)
            else:
                imm0, imm1 = instr.angle_num, instr.angle_denom
            return [
                nv.RotYInstruction(
                    lineno=instr.lineno, reg=instr.reg, imm0=imm0, imm1=imm1
                ),
            ]
        else:
            raise ValueError(
                f"Don't know how to map instruction {instr} of type {type(instr)}"
            )


def get_hardware_num_denom(
    instr: core.RotationInstruction,
) -> Tuple[Immediate, Immediate]:
    if instr.angle_denom.value not in [0, 1, 2, 3, 4]:
        raise ValueError(
            f"Instruction {instr} not supported: angle_denom is {instr.angle_denom}."
        )

    denom_diff = 4 - instr.angle_denom.value
    angle_num = instr.angle_num.value * (2**denom_diff)
    return (Immediate(angle_num), Immediate(4))



from dataclasses import dataclass

import numpy as np

from netqasm.util.quantum_gates import get_rotation_matrix

from . import core

# Explicit instruction types in the Vanilla flavour.


@dataclass
class GateXInstruction(core.SingleQubitInstruction):
    id: int = 20
    mnemonic: str = "x"

    def to_matrix(self) -> np.ndarray:
        return np.array([[0, 1], [1, 0]])


@dataclass
class GateYInstruction(core.SingleQubitInstruction):
    id: int = 21
    mnemonic: str = "y"

    def to_matrix(self) -> np.ndarray:
        return np.array([[0, -1j], [1j, 0]])


@dataclass
class GateZInstruction(core.SingleQubitInstruction):
    id: int = 22
    mnemonic: str = "z"

    def to_matrix(self) -> np.ndarray:
        return np.array([[1, 0], [0, -1]])


@dataclass
class GateHInstruction(core.SingleQubitInstruction):
    id: int = 23
    mnemonic: str = "h"

    def to_matrix(self) -> np.ndarray:
        X = GateXInstruction().to_matrix()
        Z = GateZInstruction().to_matrix()
        return (X + Z) / np.sqrt(2)  # type: ignore


@dataclass
class GateSInstruction(core.SingleQubitInstruction):
    id: int = 24
    mnemonic: str = "s"

    def to_matrix(self) -> np.ndarray:
        return np.array([[1, 0], [0, 1j]])


@dataclass
class GateKInstruction(core.SingleQubitInstruction):
    id: int = 25
    mnemonic: str = "k"

    def to_matrix(self) -> np.ndarray:
        Y = GateYInstruction().to_matrix()
        Z = GateZInstruction().to_matrix()
        return (Y + Z) / np.sqrt(2)  # type: ignore


@dataclass
class GateTInstruction(core.SingleQubitInstruction):
    id: int = 26
    mnemonic: str = "t"

    def to_matrix(self) -> np.ndarray:
        return np.array([[1, 0], [0, (1 + 1j) / np.sqrt(2)]])


@dataclass
class RotXInstruction(core.RotationInstruction):
    id: int = 27
    mnemonic: str = "rot_x"

    def to_matrix(self) -> np.ndarray:
        axis = [1, 0, 0]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_rotation_matrix(axis, angle)


@dataclass
class RotYInstruction(core.RotationInstruction):
    id: int = 28
    mnemonic: str = "rot_y"

    def to_matrix(self) -> np.ndarray:
        axis = [0, 1, 0]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_rotation_matrix(axis, angle)


@dataclass
class RotZInstruction(core.RotationInstruction):
    id: int = 29
    mnemonic: str = "rot_z"

    def to_matrix(self) -> np.ndarray:
        axis = [0, 0, 1]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_rotation_matrix(axis, angle)


@dataclass
class CnotInstruction(core.TwoQubitInstruction):
    id: int = 30
    mnemonic: str = "cnot"

    def to_matrix(self) -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])

    def to_matrix_target_only(self) -> np.ndarray:
        return np.array([[0, 1], [1, 0]])


@dataclass
class CphaseInstruction(core.TwoQubitInstruction):
    id: int = 31
    mnemonic: str = "cphase"

    def to_matrix(self) -> np.ndarray:
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])

    def to_matrix_target_only(self) -> np.ndarray:
        return np.array([[1, 0], [0, -1]])


@dataclass
class MovInstruction(core.TwoQubitInstruction):
    """Move source qubit to target qubit (target is overwritten)"""

    id: int = 41
    mnemonic: str = "mov"

    def to_matrix(self) -> np.ndarray:
        # NOTE: Currently this is represented as a full SWAP.
        return np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    def to_matrix_target_only(self) -> np.ndarray:  # type: ignore
        # NOTE: The mov instruction is not meant to be viewed as control-target gate.
        # Therefore, it is OK to not explicitly define a matrix.
        return None  # type: ignore



class NetQASMSyntaxError(SyntaxError):
    pass


class NetQASMInstrError(ValueError):
    pass


class NoCircuitRuleError(RuntimeError):
    pass


class NotAllocatedError(RuntimeError):
    pass


class SubroutineAbortedError(RuntimeError):
    pass



from netqasm.logging.output import get_new_app_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection


def measure_basis_0(alice, q):
    return q.measure()


def measure_basis_1(alice, q):
    q.H()
    return q.measure()


def main(app_config=None, x=0):
    if not (x == 0 or x == 1):
        raise ValueError(f"x should be 0 or 1, not {x}")

    app_logger = get_new_app_logger(
        app_name=app_config.app_name, log_config=app_config.log_config
    )
    app_logger.log(f"Alice received input bit x = {x}")

    epr_socket = EPRSocket("repeater")

    alice = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )

    with alice:
        # Create EPR pair with Bob, through the repeater.
        epr = epr_socket.create_keep()[0]

        # CHSH strategy: measure in one of 2 bases depending on x.
        if x == 0:
            a = measure_basis_0(alice, epr)
        else:
            a = measure_basis_1(alice, epr)

    app_logger.log(f"Alice outputs a = {a}")
    return {"a": int(a)}


if __name__ == "__main__":
    main()



import pytest

from netqasm.lang import instr as instructions
from netqasm.lang.encoding import RegisterName
from netqasm.lang.operand import Address, ArrayEntry, Immediate, Register
from netqasm.lang.parsing import parse_text_subroutine
from netqasm.lang.subroutine import Subroutine
from netqasm.lang.version import NETQASM_VERSION
from netqasm.util.error import NetQASMInstrError, NetQASMSyntaxError


@pytest.mark.parametrize(
    "subroutine, error",
    [
        ("# WRONG", NetQASMInstrError),  # Wrong keyword
        ("# APPID 0\nH 0\n#APPID 0", NetQASMSyntaxError),  # Preamble after body
        ("# DEFINE args {0, 0", NetQASMSyntaxError),  # No end-bracket
        ("# NETQASM", NetQASMSyntaxError),  # No argument
        ("# NETQASM 1 2", NetQASMSyntaxError),  # Two arguments
        ("# APPID", NetQASMSyntaxError),  # No argument
        ("# APPID 1 2", NetQASMSyntaxError),  # Two arguments
        ("# DEFINE args", NetQASMSyntaxError),  # One arguments
        ("# DEFINE args 0 0", NetQASMSyntaxError),  # Three arguments
        ("# DEFINE 1args 0", NetQASMInstrError),  # Not a valid macro key
        (
            "# DEFINE args 0\n# DEFINE args 1",
            NetQASMInstrError,
        ),  # Not unique macro keys
    ],
)
def test_faulty_preamble(subroutine, error):
    with pytest.raises(error):
        parse_text_subroutine(subroutine)


def test_simple():
    subroutine = """
# NETQASM 0.0
# APPID 0
set R0 0
store R0 @0[R2]
set Q0 0
init Q0
array(4) @2
add R1 R2 1
beq 0 0 EXIT
EXIT:
ret_reg R0
ret_arr @0
"""

    expected = Subroutine(
        netqasm_version=NETQASM_VERSION,
        app_id=0,
        instructions=[
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 0),
                imm=Immediate(0),
            ),
            instructions.core.StoreInstruction(
                reg=Register(RegisterName.R, 0),
                entry=ArrayEntry(Address(0), Register(RegisterName.R, 2)),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(0),
            ),
            instructions.core.InitInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 3),
                imm=Immediate(4),
            ),
            instructions.core.ArrayInstruction(
                reg=Register(RegisterName.R, 3),
                address=Address(address=2),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 3),
                imm=Immediate(1),
            ),
            instructions.core.AddInstruction(
                reg0=Register(RegisterName.R, 1),
                reg1=Register(RegisterName.R, 2),
                reg2=Register(RegisterName.R, 3),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 3),
                imm=Immediate(0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 4),
                imm=Immediate(0),
            ),
            instructions.core.BeqInstruction(
                reg0=Register(RegisterName.R, 3),
                reg1=Register(RegisterName.R, 4),
                imm=Immediate(11),
            ),
            instructions.core.RetRegInstruction(
                reg=Register(RegisterName.R, 0),
            ),
            instructions.core.RetArrInstruction(
                address=Address(0),
            ),
        ],
    )

    subroutine = parse_text_subroutine(subroutine)
    print(subroutine)
    for i, instr in enumerate(subroutine.instructions):
        exp_instr = expected.instructions[i]
        print(repr(instr))
        print(repr(exp_instr))
        assert instr == exp_instr
    print(repr(subroutine))
    print(repr(expected))


def test_loop():
    subroutine = """
# NETQASM 0.0
# APPID 0
# DEFINE ms @0
// Setup constants
set C1 1
set C10 10
// Setup classical registers
set Q0 0
set R0 0
array C10 $ms

// Loop entry
LOOP:
beq R0 C10 EXIT

// Loop body
qalloc Q0
init Q0
h Q0
meas Q0 M0

// Store to array
store M0 $ms[R0]

qfree Q0
add R0 R0 C1

// Loop exit
jmp LOOP
EXIT:
"""

    expected = Subroutine(
        netqasm_version=NETQASM_VERSION,
        app_id=0,
        instructions=[
            instructions.core.SetInstruction(
                reg=Register(RegisterName.C, 1),
                imm=Immediate(1),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.C, 10),
                imm=Immediate(10),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(0),
            ),
            instructions.core.SetInstruction(
                reg=Register(RegisterName.R, 0),
                imm=Immediate(0),
            ),
            instructions.core.ArrayInstruction(
                reg=Register(RegisterName.C, 10),
                address=Address(0),
            ),
            instructions.core.BeqInstruction(
                reg0=Register(RegisterName.R, 0),
                reg1=Register(RegisterName.C, 10),
                imm=Immediate(14),
            ),
            instructions.core.QAllocInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.InitInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.vanilla.GateHInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.MeasInstruction(
                reg0=Register(RegisterName.Q, 0),
                reg1=Register(RegisterName.M, 0),
            ),
            instructions.core.StoreInstruction(
                reg=Register(RegisterName.M, 0),
                entry=ArrayEntry(
                    address=Address(0),
                    index=Register(RegisterName.R, 0),
                ),
            ),
            instructions.core.QFreeInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.AddInstruction(
                reg0=Register(RegisterName.R, 0),
                reg1=Register(RegisterName.R, 0),
                reg2=Register(RegisterName.C, 1),
            ),
            instructions.core.JmpInstruction(
                imm=Immediate(5),
            ),
        ],
    )
    subroutine = parse_text_subroutine(subroutine)
    for i, instr in enumerate(subroutine.instructions):
        exp_instr = expected.instructions[i]
        print(repr(instr))
        print(repr(exp_instr))
        assert instr == exp_instr
    print(repr(subroutine))
    print(repr(expected))


def test_rotations():
    subroutine = """
# NETQASM 0.0
# APPID 0

set Q0 0
qalloc Q0
init Q0

// Perform rotations
rot_x Q0 1 1  // rotate by 1 * pi / 1 = pi
rot_x Q0 1 4  // rotate by 1 * pi / 4 = pi / 4
rot_y Q0 7 22  // rotate by 7 pi / 22

qfree Q0
"""

    expected = Subroutine(
        netqasm_version=NETQASM_VERSION,
        app_id=0,
        instructions=[
            instructions.core.SetInstruction(
                reg=Register(RegisterName.Q, 0),
                imm=Immediate(0),
            ),
            instructions.core.QAllocInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            instructions.core.InitInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
            # Rotations
            instructions.vanilla.RotXInstruction(
                reg=Register(RegisterName.Q, 0),
                imm0=Immediate(1),
                imm1=Immediate(1),
            ),
            instructions.vanilla.RotXInstruction(
                reg=Register(RegisterName.Q, 0),
                imm0=Immediate(1),
                imm1=Immediate(4),
            ),
            instructions.vanilla.RotYInstruction(
                reg=Register(RegisterName.Q, 0),
                imm0=Immediate(7),
                imm1=Immediate(22),
            ),
            instructions.core.QFreeInstruction(
                reg=Register(RegisterName.Q, 0),
            ),
        ],
    )
    subroutine = parse_text_subroutine(subroutine)
    for i, instr in enumerate(subroutine.instructions):
        exp_instr = expected.instructions[i]
        print(repr(instr))
        print(repr(exp_instr))
        assert instr == exp_instr
    print(repr(subroutine))
    print(repr(expected))


if __name__ == "__main__":
    test_simple()
    test_loop()
    test_rotations()



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Callable, List, Union

from netqasm.qlink_compat import LinkLayerOKTypeK, LinkLayerOKTypeM, LinkLayerOKTypeR
from netqasm.sdk.futures import Future, RegFuture
from netqasm.sdk.qubit import FutureQubit

if TYPE_CHECKING:
    from netqasm.sdk import connection
    from netqasm.sdk.builder import Builder

T_LinkLayerOkList = Union[
    List[LinkLayerOKTypeK], List[LinkLayerOKTypeM], List[LinkLayerOKTypeR]
]
T_PostRoutine = Callable[["Builder", Union[FutureQubit, List[Future]], RegFuture], None]

# Callback function type for conditional statements.
T_BranchRoutine = Callable[["connection.BaseNetQASMConnection"], None]

# Callback function type for loop statements.
T_LoopRoutine = Callable[["connection.BaseNetQASMConnection", RegFuture], None]

# Function that is called when the exit condition of a while-loop is False.
T_CleanupRoutine = Callable[["connection.BaseNetQASMConnection"], None]


class HardwareConfig(ABC):
    """Base class for hardware information used by the Builder."""

    def __init__(self, num_comm_qubits: int, num_mem_qubits: int) -> None:
        assert num_comm_qubits > 0
        assert num_mem_qubits >= 0
        self._comm_qubit_count = num_comm_qubits
        self._mem_qubit_count = num_mem_qubits

    @property
    def comm_qubit_count(self) -> int:
        return self._comm_qubit_count

    @property
    def mem_qubit_count(self) -> int:
        return self._mem_qubit_count

    @property
    def qubit_count(self) -> int:
        """Total number of qubits available."""
        return self.comm_qubit_count + self.mem_qubit_count


class NVHardwareConfig(HardwareConfig):
    """Hardware information for NV, where there is only one communication qubit."""

    def __init__(self, num_qubits: int) -> None:
        super().__init__(1, num_qubits - 1)


class GenericHardwareConfig(HardwareConfig):
    """Hardware information for a generic platform where all qubits are
    communication qubits."""

    def __init__(self, num_qubits: int) -> None:
        super().__init__(num_qubits, 0)



from netqasm.lang.instr.vanilla import CphaseInstruction
from netqasm.lang.parsing import deserialize, parse_text_subroutine


def test():
    subroutine = """
# NETQASM 0.0
# APPID 0
# DEFINE ms @0

// Setup classical registers
set Q0 0
array 10 $ms
set R0 0

// Loop entry
LOOP:
beq R0 10 EXIT

// Loop body
qalloc Q0
init Q0
h Q0
meas Q0 M0

// Store to array
store M0 $ms[R0]

qfree Q0
add R0 R0 1

// Loop exit
beq 0 0 LOOP
EXIT:
ret_reg M0
    """

    subroutine = parse_text_subroutine(subroutine)
    print(subroutine)
    data = bytes(subroutine)
    print(data)

    parsed_subroutine = deserialize(data)
    print(parsed_subroutine)

    for instr, parsed_instr in zip(
        subroutine.instructions, parsed_subroutine.instructions
    ):
        print()
        print(repr(instr))
        print(repr(parsed_instr))
        assert instr == parsed_instr


def test_rotations():
    subroutine = """
# NETQASM 0.0
# APPID 0

set Q0 0
qalloc Q0
init Q0

// Perform rotations
rot_x Q0 1 1  // rotate by 1 * pi / 1 = pi
rot_x Q0 1 4  // rotate by 1 * pi / 4 = pi / 4
rot_y Q0 7 22  // rotate by 7 pi / 22

qfree Q0
"""

    subroutine = parse_text_subroutine(subroutine)
    print(subroutine)
    data = bytes(subroutine)
    print(f"binary subroutine: {data}")

    parsed_subroutine = deserialize(data)
    print(parsed_subroutine)

    for instr, parsed_instr in zip(
        subroutine.instructions, parsed_subroutine.instructions
    ):
        print(f"instr: {instr}, parsed_instr: {parsed_instr}")
        assert instr == parsed_instr


def test_deserialize_subroutine():
    metadata = b"\x00\x00\x00\x00"
    cphase_gate = b"\x1F\x00\x00\x00\x00\x00\x00"
    raw = bytes(metadata + cphase_gate)
    print(raw)
    subroutine = deserialize(raw)
    print(subroutine)
    for instr in subroutine.instructions:
        if isinstance(instr, CphaseInstruction):
            print(f"reg0: {instr.reg0}, reg1: {instr.reg1}")

    subroutine2 = deserialize(raw)
    print(subroutine2)


if __name__ == "__main__":
    test()
    test_rotations()
    test_deserialize_subroutine()



import inspect
import os
from typing import Optional


class HostLine:
    def __init__(self, filename, lineno):
        self.filename = filename
        self.lineno = lineno

    def __str__(self):
        return str(self.lineno)


class LineTracker:
    def __init__(self, log_config):
        """
        Parameters
        ----------
        log_config : :class:`~.sdk.config.LogConfig`
        """
        self._track_lines = log_config.track_lines
        if not self._track_lines:
            return
        if log_config.app_dir is None:
            raise RuntimeError("Cannot create Linetracker because app_dir is None")

        self.app_dir = os.path.abspath(log_config.app_dir)

        lib_dirs = log_config.lib_dirs
        if lib_dirs is None:
            lib_dirs = []
        self.lib_dirs = [os.path.abspath(dir) for dir in lib_dirs]

    def _get_file_from_frame(self, frame):
        return os.path.abspath(frame.f_code.co_filename)

    def get_line(self) -> Optional[HostLine]:
        if not self._track_lines:
            return None

        frame = inspect.currentframe()
        assert frame is not None

        frame_found = False
        while not frame_found:
            frame_file = self._get_file_from_frame(frame)

            # first check if it's coming from one of the lib directories
            for lib_dir in self.lib_dirs:
                if frame_file.startswith(lib_dir):
                    frame_found = True
                    break

            # check in app directory itself
            if frame_file.startswith(self.app_dir):
                frame_found = True

            if frame_found:
                break

            frame = frame.f_back
            if frame is None:
                raise RuntimeError(f"No frame found in directory {self.app_dir}")

        filename = os.path.abspath(frame.f_code.co_filename)
        return HostLine(filename, frame.f_lineno)



import abc
import os
from dataclasses import asdict
from datetime import datetime
from enum import Enum
from typing import List, Optional, Set, Tuple

from netqasm.lang import instr as instructions
from netqasm.lang.encoding import RegisterName
from netqasm.lang.operand import Address, ArrayEntry, Register
from netqasm.qlink_compat import RequestType
from netqasm.runtime.interface.logging import (
    AppLogEntry,
    ClassCommLogEntry,
    EntanglementType,
    InstrLogEntry,
    NetworkLogEntry,
    QubitGroups,
    QubitState,
)
from netqasm.util.error import NotAllocatedError
from netqasm.util.log import LineTracker
from netqasm.util.yaml import dump_yaml


def should_ignore_instr(instr):
    return (
        isinstance(instr, instructions.core.SetInstruction)
        or isinstance(instr, instructions.core.QAllocInstruction)
        or isinstance(instr, instructions.core.QFreeInstruction)
        or isinstance(instr, instructions.core.CreateEPRInstruction)
        or isinstance(instr, instructions.core.RecvEPRInstruction)
    )


# Keep track of all structured loggers
# to be able to save them while finished applications
_STRUCT_LOGGERS: List["StructuredLogger"] = []


def reset_struct_loggers():
    while len(_STRUCT_LOGGERS) > 0:
        _STRUCT_LOGGERS.pop()


def save_all_struct_loggers():
    while len(_STRUCT_LOGGERS) > 0:
        struct_logger = _STRUCT_LOGGERS.pop()
        struct_logger.save()


class StructuredLogger(abc.ABC):
    def __init__(self, filepath):
        self._filepath = filepath

        self._storage = []

        _STRUCT_LOGGERS.append(self)

    def log(self, *args, **kwargs):
        entry = self._construct_entry(*args, **kwargs)
        if entry is not None:
            self._storage.append(entry)

    @abc.abstractmethod
    def _construct_entry(self, *args, **kwargs):
        pass

    def _get_op_values(self, subroutine_id, operands):
        values = []
        for operand in operands:
            value = self._get_op_value(
                subroutine_id=subroutine_id,
                operand=operand,
            )
            values.append(value)
        return values

    def _get_op_value(self, subroutine_id, operand):
        app_id = self._executor._get_app_id(subroutine_id=subroutine_id)
        value = None
        if isinstance(operand, int):
            value = operand
        elif isinstance(operand, Register):
            value = self._executor._get_register(app_id=app_id, register=operand)
        elif isinstance(operand, ArrayEntry):
            value = self._executor._get_array_entry(app_id=app_id, array_entry=operand)
        return value

    def save(self):
        dump_yaml(self._storage, self._filepath)


class InstrLogger(StructuredLogger):

    # Absulute IDs of all qubits in the simulation
    # List of (node_name, subroutine_id, qubit_id)
    _qubits: Set[Tuple[str, int, int]] = set()

    def __init__(self, filepath: str, executor):
        super().__init__(filepath)
        self._executor = executor

    def _construct_entry(self, *args, **kwargs):
        command = kwargs["command"]
        app_id = kwargs["app_id"]
        subroutine_id = kwargs["subroutine_id"]
        output = kwargs["output"]
        wall_time = str(datetime.now())
        sim_time = self._executor._get_simulated_time()
        program_counter = kwargs["program_counter"]
        instr_name = command.mnemonic
        operands = command.operands
        op_values = self._get_op_values(subroutine_id=subroutine_id, operands=operands)
        ops_str = [f"{op}={opv}" for op, opv in zip(operands, op_values)]
        log = f"Doing instruction {instr_name} with operands {ops_str}"
        virtual_qubit_ids, physical_qubit_ids = self._get_qubit_ids(
            subroutine_id=subroutine_id,
            command=command,
        )
        self._update_qubits(
            subroutine_id=subroutine_id,
            instr=command,
            qubit_ids=virtual_qubit_ids,
        )
        if should_ignore_instr(command):
            return None
        if len(virtual_qubit_ids) == 0:
            # Not a qubit instruction
            return None
        if isinstance(command, instructions.core.MeasInstruction):
            outcome = output
        else:
            outcome = None
        if isinstance(command, instructions.core.RotationInstruction) or isinstance(
            command, instructions.core.ControlledRotationInstruction
        ):
            num = command.angle_num.value
            denom = 2**command.angle_denom.value
            angle = {"num": num, "den": denom}
        else:
            angle = None

        qubit_groups = self._get_qubit_groups()
        return asdict(
            InstrLogEntry(
                WCT=wall_time,
                SIT=sim_time,
                AID=app_id,
                SID=subroutine_id,
                PRC=program_counter,
                HLN=None,
                HFL=None,
                INS=instr_name,
                OPR=ops_str,
                ANG=angle,
                QID=virtual_qubit_ids,
                VID=physical_qubit_ids,
                OUT=outcome,
                QGR=qubit_groups,
                LOG=log,
            )
        )

    def _get_qubit_ids(
        self,
        subroutine_id: int,
        command: instructions.base.NetQASMInstruction,
    ) -> Tuple[List[int], List[int]]:
        """Gets the qubit IDs involved in a command"""
        # If EPR then get the qubit IDs from the array
        epr_instructions = [
            instructions.core.CreateEPRInstruction,
            instructions.core.RecvEPRInstruction,
        ]
        app_id = self._executor._get_app_id(subroutine_id=subroutine_id)
        if any(isinstance(command, cmd_cls) for cmd_cls in epr_instructions):
            # Ignore a constant register since this indicates it's a measure directly request
            if command.qubit_addr_array.name == RegisterName.C:  # type: ignore
                return [], []
            qubit_id_array_address = Address(
                self._get_op_value(
                    subroutine_id=subroutine_id,
                    operand=command.qubit_addr_array,  # type: ignore
                )
            )
            virtual_qubit_ids = self._executor._get_array(
                app_id=app_id,
                address=qubit_id_array_address,
            )  # type: ignore
        else:
            # Otherwise just get the qubits from the operands which have name Q
            virtual_qubit_ids = []
            for operand in command.operands:
                if isinstance(operand, Register) and operand.name == RegisterName.Q:
                    virtual_qubit_id = self._get_op_value(
                        subroutine_id=subroutine_id,
                        operand=operand,
                    )
                    virtual_qubit_ids.append(virtual_qubit_id)

        # Lookup physical qubit IDs
        physical_qubit_ids = self._get_physical_qubit_ids(
            app_id=app_id,
            virtual_qubit_ids=virtual_qubit_ids,
        )
        return virtual_qubit_ids, physical_qubit_ids

    def _get_physical_qubit_ids(self, app_id, virtual_qubit_ids):
        physical_qubit_ids = []
        for virtual_qubit_id in virtual_qubit_ids:
            try:
                physical_qubit_id = self._executor._get_position_in_unit_module(
                    app_id=app_id,
                    address=virtual_qubit_id,
                )
            except NotAllocatedError:
                physical_qubit_id = None
            physical_qubit_ids.append(physical_qubit_id)

        return physical_qubit_ids

    def _update_qubits(
        self,
        subroutine_id: int,
        instr: instructions.base.NetQASMInstruction,
        qubit_ids: List[int],
    ) -> None:
        add_qubit_instrs = [
            instructions.core.InitInstruction,
            instructions.core.CreateEPRInstruction,
            instructions.core.RecvEPRInstruction,
        ]
        remove_qubit_instrs = [
            instructions.core.QFreeInstruction,
            instructions.core.MeasInstruction,
        ]
        node_name = self._get_node_name()
        app_id = self._get_app_id(subroutine_id=subroutine_id)
        if any(isinstance(instr, cmd_cls) for cmd_cls in add_qubit_instrs):
            for qubit_id in qubit_ids:
                abs_id = node_name, app_id, qubit_id
                self.__class__._qubits.add(abs_id)
        elif any(isinstance(instr, cmd_cls) for cmd_cls in remove_qubit_instrs):
            for qubit_id in qubit_ids:
                abs_id = node_name, app_id, qubit_id
                if abs_id in self.__class__._qubits:
                    self.__class__._qubits.remove(abs_id)

    def _get_app_id(self, subroutine_id: int) -> int:
        """Returns the app ID for a given subroutine ID"""
        return self._executor._get_app_id(subroutine_id=subroutine_id)  # type: ignore

    @classmethod
    def _get_qubit_states(
        cls,
        subroutine_id: int,
        qubit_ids: List[int],
    ) -> Optional[List[QubitState]]:
        """Returns the reduced qubit states of the qubits involved in a command"""
        # NOTE should be subclassed
        raise NotImplementedError

    @classmethod
    def _get_qubit_groups(cls) -> Optional[QubitGroups]:
        """Returns the current qubit groups in the simulation (qubits which have interacted
        and therefore may or may not be entangled)"""
        # NOTE should be subclassed
        raise NotImplementedError

    def _get_node_name(self) -> str:
        # NOTE should be subclassed
        raise NotImplementedError


class NetworkLogger(StructuredLogger):
    def __init__(self, filepath):
        super().__init__(filepath)

    def _construct_entry(self, *args, **kwargs):
        wall_time = str(datetime.now())
        sim_time = kwargs["sim_time"]
        ent_type = kwargs["ent_type"]
        if ent_type == RequestType.M:
            ent_type = f"epr_{EntanglementType.MD}"
        elif ent_type == RequestType.K:
            ent_type = f"epr_{EntanglementType.CK}"
        meas_bases = kwargs["meas_bases"]
        meas_outcomes = kwargs["meas_outcomes"]

        ent_stage = kwargs["ent_stage"]
        nodes = kwargs["nodes"]
        path = kwargs["path"]
        qubit_ids = kwargs["qubit_ids"]
        qubit_groups = kwargs["qubit_groups"]
        msg = kwargs["msg"]
        return asdict(
            NetworkLogEntry(
                WCT=wall_time,
                SIT=sim_time,
                TYP=ent_type,
                INS=f"epr_{ent_stage}",
                BAS=meas_bases,
                MSR=meas_outcomes,
                NOD=nodes,
                PTH=path,
                QID=qubit_ids,
                QGR=qubit_groups,
                LOG=msg,
            )
        )


class SocketOperation(Enum):
    SEND = "SEND"
    RECV = "RECV"
    WAIT_RECV = "WAIT_RECV"


class ClassCommLogger(StructuredLogger):
    def _construct_entry(self, *args, **kwargs):
        socket_op = kwargs["socket_op"]
        msg = kwargs["msg"]
        sender = kwargs["sender"]
        receiver = kwargs["receiver"]
        socket_id = kwargs["socket_id"]
        hln = kwargs["hln"]
        hfl = kwargs["hfl"]
        log = kwargs["log"]
        wall_time = str(datetime.now())
        return asdict(
            ClassCommLogEntry(
                WCT=wall_time,
                HLN=hln,
                HFL=hfl,
                INS=socket_op.value,
                MSG=msg,
                SEN=sender,
                REC=receiver,
                SOD=socket_id,
                LOG=log,
            )
        )


class AppLogger(StructuredLogger):
    def __init__(self, filepath, log_config):
        super().__init__(filepath=filepath)

        self._line_tracker = LineTracker(log_config)

    def _construct_entry(self, *args, **kwargs):
        log = kwargs.get("log", None)
        if log is None:
            assert len(args) == 1, "AppLogger only takes on argument"
            log = args[0]
        # host_line = self._line_tracker.get_line()
        hln = None  # TODO: fix
        hfl = None  # TODO: fix
        wall_time = str(datetime.now())
        return asdict(AppLogEntry(WCT=wall_time, HLN=hln, HFL=hfl, LOG=log))


def get_new_app_logger(app_name, log_config):
    filename = f"{str(app_name).lower()}_app_log.yaml"
    log_dir = log_config.log_subroutines_dir
    filepath = os.path.join(log_dir, filename)
    app_logger = AppLogger(filepath=filepath, log_config=log_config)
    return app_logger



from netqasm.sdk import EPRSocket, Qubit
from netqasm.sdk.external import NetQASMConnection, Socket, get_qubit_state
from netqasm.sdk.toolbox import set_qubit_state


def main(app_config=None, phi=0.0, theta=0.0):
    # socket for creating an EPR pair with target
    target_epr = EPRSocket("target")

    # socket for communicating classical messages with target
    class_socket = Socket("controller", "target", log_config=app_config.log_config)

    # connect to the back-end
    controller = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[target_epr],
    )

    with controller:
        # create one EPR pair with target
        epr = target_epr.create_keep(1)[0]

        # initialize control qubit of the distributed CNOT
        ctrl_qubit = Qubit(controller)
        set_qubit_state(ctrl_qubit, phi, theta)

        # perform a local CNOT with `epr` and measure `epr`
        ctrl_qubit.cnot(epr)
        epr_meas = epr.measure()

        # let back-end execute the quantum operations above
        controller.flush()

        # send the outcome to target
        class_socket.send(str(epr_meas))

        # wait for target's measurement outcome to undo potential entanglement
        # between his EPR half and the original control qubit
        target_meas = class_socket.recv()
        if target_meas == "1":
            ctrl_qubit.Z()

        controller.flush()

        # ack the outcome
        class_socket.send("ACK")

        # get the combined state of Controller's control and target's target
        dm = get_qubit_state(ctrl_qubit, reduced_dm=False)

        return {
            "epr_meas": int(epr_meas),
            "final_state": dm if dm is None else dm.tolist(),
        }



"""Global management for classical sockets that live in separate threads.

This module contains the _SocketHub that manages sockets that are in separate threads
and need to communicate with each other.
"""

from __future__ import annotations

from collections import defaultdict
from threading import Lock
from time import sleep
from timeit import default_timer as timer
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Union
from weakref import WeakMethod

from netqasm.logging.glob import get_netqasm_logger

if TYPE_CHECKING:
    import logging

    from netqasm.sdk.classical_communication import message, thread_socket


class _SocketHub:
    """Global manager for classical sockets that live in separate threads.

    This class is used by ThreadSockets and is typically not used directly.
    """

    _CONNECT_SLEEP_TIME: float = 0.1
    _RECV_SLEEP_TIME: float = 0.1

    def __init__(self):
        """Used to connect all sockets (:class:`~.ThreadSocket`) used between threads"""
        # NOTE a socket gets tracked at two places, the _open_sockets and _remote_sockets
        # the idea is that when a connection is closed it removes itself from the _open_sockets
        # but it is the task of the other end to remote it from the _remote_sockets.
        # This is to prevent a connection from opening and closing on one side before
        # the sides notices and waits for ever.
        self._open_sockets: Set[thread_socket.socket.T_ThreadSocketKey] = set()
        self._remote_sockets: Set[thread_socket.socket.T_ThreadSocketKey] = set()

        self._messages: Dict[
            thread_socket.socket.T_ThreadSocketKey,
            List[Union[str, message.StructuredMessage]],
        ] = defaultdict(list)
        self._recv_callbacks: Dict[
            thread_socket.socket.T_ThreadSocketKey, WeakMethod
        ] = {}
        self._conn_lost_callbacks: Dict[
            thread_socket.socket.T_ThreadSocketKey, WeakMethod
        ] = {}

        self._lock: Lock = Lock()

        self._logger: logging.Logger = get_netqasm_logger(self.__class__.__name__)

    def connect(
        self, socket: thread_socket.ThreadSocket, timeout: Optional[float] = None
    ) -> None:
        """Connects a socket to another"""
        self._open_sockets.add(socket.key)
        self._remote_sockets.add(socket.key)
        self._add_callbacks(socket)

        self._wait_for_remote(socket, timeout=timeout)

    def _add_callbacks(self, socket: thread_socket.ThreadSocket) -> None:
        if socket.use_callbacks:
            self._recv_callbacks[socket.key] = WeakMethod(socket.recv_callback)  # type: ignore
            self._conn_lost_callbacks[socket.key] = WeakMethod(socket.conn_lost_callback)  # type: ignore

    def is_connected(self, socket: thread_socket.ThreadSocket) -> bool:
        return all(key in self._open_sockets for key in [socket.key, socket.remote_key])

    def disconnect(self, socket: thread_socket.ThreadSocket) -> None:
        """Disconnect a socket"""
        with self._lock:
            conn_lost_callback = self._conn_lost_callbacks.get(socket.remote_key)
            if conn_lost_callback is not None:
                method = conn_lost_callback()
                # This is a WeakMethod so check if it exists
                if method is None:
                    self._logger.warning(
                        f"Trying to call lost connection callback "
                        f"for socket {socket.remote_key} but object is garbage collected"
                    )
                else:
                    method()

            if socket.key in self._open_sockets:
                self._open_sockets.remove(socket.key)
            if socket.remote_key in self._remote_sockets:
                self._remote_sockets.remove(socket.remote_key)
            self._recv_callbacks.pop(socket.key, None)
            self._conn_lost_callbacks.pop(socket.key, None)

    def _wait_for_remote(
        self, socket: thread_socket.ThreadSocket, timeout: Optional[float] = None
    ) -> None:
        """Wait for a remote socket to become active"""
        t_start = timer()
        while True:
            if socket.remote_key in self._open_sockets:
                self._logger.debug(f"Connection for socket {socket.key} successful")
                return
            if socket.remote_key in self._remote_sockets:
                self._logger.debug(
                    f"Connection for socket {socket.key} was successful but closed again"
                )
                return
            t_now = timer()
            t_elapsed = t_now - t_start
            if timeout is not None:
                if t_elapsed > timeout:
                    app_name = socket.app_name
                    remote_app_name = socket.remote_app_name
                    socket_id = socket.id
                    raise TimeoutError(
                        f"Timeout while connection node ID {app_name} to "
                        f"{remote_app_name} using socket {socket_id}"
                    )
            self._logger.debug(
                f"Connection for socket {socket.key} failed, "
                f"trying again in {self._CONNECT_SLEEP_TIME} s..."
            )
            sleep(self.__class__._CONNECT_SLEEP_TIME)

    def send(
        self,
        socket: thread_socket.ThreadSocket,
        msg: Union[str, message.StructuredMessage],
    ) -> None:
        """Send a message using a given socket"""
        recv_callback = self._recv_callbacks.get(socket.remote_key)
        if recv_callback is not None:
            self._logger.debug(
                f"Message {msg} sent on socket {socket.key}, calling callback for recv"
            )
            method = recv_callback()
            # This is a WeakMethod so check if it exists
            if method is None:
                self._logger.warning(
                    f"Trying to call recv callback "
                    f"for socket {socket.remote_key} but object is garbage collected"
                )
            else:
                method(msg)
        else:
            self._logger.debug(
                f"Message {msg} sent on socket {socket.key}, adding to pending received messages"
            )
            with self._lock:
                self._messages[socket.remote_key].append(msg)

    def recv(
        self,
        socket: thread_socket.ThreadSocket,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> Union[str, message.StructuredMessage]:
        """Recv a message to a given socket"""
        t_start = timer()
        while True:
            with self._lock:
                messages = self._messages[socket.key]
            if len(messages) == 0:
                if not block:
                    raise RuntimeError(f"No message to receive on socket {socket.key}")
            else:
                with self._lock:
                    msg = messages.pop(0)
                self._logger.debug(f"Got message {msg} for socket {socket.key}")
                return msg
            if timeout is not None:
                t_now = timer()
                t_elapsed = t_now - t_start
                if t_elapsed > timeout:
                    raise TimeoutError(
                        f"Timeout while trying to receive message for socket {socket.key}"
                    )
            self._logger.debug(
                f"No message yet for socket {socket.key}, "
                f"trying again in {self._RECV_SLEEP_TIME} s..."
            )
            sleep(self.__class__._RECV_SLEEP_TIME)


_socket_hub: _SocketHub = _SocketHub()


def reset_socket_hub() -> None:
    _socket_hub.__init__()  # type: ignore



# Exports not-implemented function such that the DebugConnection can be used

from .application import ApplicationInstance


def get_qubit_state(qubit, reduced_dm=True):
    return None


def run_application(
    app_instance: ApplicationInstance,
    post_function=None,
    instr_log_dir=None,
    results_file=None,
    use_app_config=True,  # whether to give app_config as argument to app's main()
):
    raise NotImplementedError("Running applications in debug mode not yet supported")



from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection

logger = get_netqasm_logger()


def main(app_config):
    epr_socket_alice = EPRSocket(
        remote_app_name="alice", epr_socket_id=0, remote_epr_socket_id=1
    )
    epr_socket_bob = EPRSocket(
        remote_app_name="bob", epr_socket_id=1, remote_epr_socket_id=1
    )

    charlie = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket_alice, epr_socket_bob],
    )
    with charlie:
        epr_alice = epr_socket_alice.recv_keep()[0]
        m_alice = epr_alice.measure()

        charlie.flush()

        epr_bob = epr_socket_bob.recv_keep()[0]
        m_bob = epr_bob.measure()

    logger.info(f"charlie:  m_alice:  {m_alice}")
    logger.info(f"charlie:  m_bob:    {m_bob}")



"""
Quantum node controller interface for simulators.

This module provides the `QNodeController` class which can be used by simulators
as a base class for modeling the quantum node controller.
"""

import abc
import logging
from types import GeneratorType
from typing import Any, Callable, Dict, Generator, List, Optional, Set, Type

from netqasm.backend.executor import Executor
from netqasm.backend.messages import (
    InitNewAppMessage,
    Message,
    MessageType,
    OpenEPRSocketMessage,
    Signal,
    SignalMessage,
    StopAppMessage,
    SubroutineMessage,
)
from netqasm.backend.network_stack import BaseNetworkStack
from netqasm.lang.instr import Flavour
from netqasm.lang.parsing import deserialize
from netqasm.lang.subroutine import Subroutine
from netqasm.logging.glob import get_netqasm_logger


class QNodeController:
    """Class for representing a Quantum Node Controller in a simulation.

    A QNodeController represents a physical quantum node controller that handles
    messages coming from the Host, lets the Executor execute subroutines, and sends
    results back to the Host.
    """

    def __init__(
        self,
        name: str,
        instr_log_dir: Optional[str] = None,
        flavour: Optional[Flavour] = None,
        **kwargs,
    ) -> None:
        """QNodeController constructor.

        :param name: name used for logging purposes
        :param instr_log_dir: directory used to write log files to
        :param flavour: which NetQASM flavour this quantum node controller should
            expect and be able to interpret
        """
        self.name: str = name

        self.flavour: Optional[Flavour] = flavour

        self._executor: Executor = self._get_executor_class(flavour=flavour)(
            name=name,
            instr_log_dir=instr_log_dir,
            **kwargs,
        )

        self._message_handlers: Dict[
            MessageType, Callable
        ] = self._get_message_handlers()

        # Keep track of active apps
        self._active_app_ids: Set[int] = set()

        # Keep track of finished messages
        self._finished_messages: List[bytes] = []

        self._finished: bool = False

        self._logger: logging.Logger = get_netqasm_logger(
            f"{self.__class__.__name__}({self.name})"
        )

    @classmethod
    @abc.abstractmethod
    def _get_executor_class(cls, flavour: Optional[Flavour] = None) -> Type[Executor]:
        pass

    @abc.abstractmethod
    def stop(self) -> None:
        pass

    @property
    def finished(self) -> bool:
        return self._finished

    def handle_netqasm_message(
        self, msg_id: int, msg: Message
    ) -> Generator[Any, None, None]:
        yield from self._handle_message(msg_id=msg_id, msg=msg)

    def _handle_message(self, msg_id: int, msg: Message) -> Generator[Any, None, None]:
        self._logger.info(f"Handle message {msg}")
        output = self._message_handlers[msg.TYPE](msg)
        if isinstance(output, GeneratorType):
            yield from output
        self._mark_message_finished(msg_id=msg_id, msg=msg)
        if self.finished:
            self.stop()

    @property
    def has_active_apps(self) -> bool:
        return len(self._active_app_ids) > 0

    @property
    def network_stack(self) -> Optional[BaseNetworkStack]:
        return self._executor.network_stack

    @network_stack.setter
    def network_stack(self, network_stack: BaseNetworkStack) -> None:
        self._executor.network_stack = network_stack

    def _get_message_handlers(self) -> Dict[MessageType, Callable]:
        return {
            MessageType.SIGNAL: self._handle_signal,
            MessageType.SUBROUTINE: self._handle_subroutine,
            MessageType.INIT_NEW_APP: self._handle_init_new_app,
            MessageType.STOP_APP: self._handle_stop_app,
            MessageType.OPEN_EPR_SOCKET: self._handle_open_epr_socket,
        }

    def add_network_stack(self, network_stack: BaseNetworkStack) -> None:
        self._executor.network_stack = network_stack

    @abc.abstractmethod
    def _mark_message_finished(self, msg_id: int, msg: Message) -> None:
        pass

    def _handle_subroutine(self, msg: SubroutineMessage) -> Generator[Any, None, None]:
        subroutine = deserialize(msg.subroutine, flavour=self.flavour)
        self._logger.debug(
            f"Executing next subroutine " f"from app ID {subroutine.app_id}"
        )
        yield from self._execute_subroutine(subroutine=subroutine)

    def _execute_subroutine(self, subroutine: Subroutine) -> Generator[Any, None, None]:
        yield from self._executor.execute_subroutine(subroutine=subroutine)

    def _handle_init_new_app(self, msg: InitNewAppMessage) -> None:
        app_id = msg.app_id
        self._add_app(app_id=app_id)
        max_qubits = msg.max_qubits
        self._logger.debug(
            f"Allocating a new "
            f"unit module of size {max_qubits} for application with app ID {app_id}.\n"
        )
        self._executor.init_new_application(
            app_id=app_id,
            max_qubits=max_qubits,
        )

    def _add_app(self, app_id: int) -> None:
        self._active_app_ids.add(app_id)

    def _remove_app(self, app_id: int) -> None:
        self._active_app_ids.remove(app_id)

    def _handle_stop_app(self, msg: StopAppMessage) -> Generator[Any, None, None]:
        app_id = msg.app_id
        self._remove_app(app_id=app_id)
        self._logger.debug(f"Stopping application with app ID {app_id}")
        yield from self._executor.stop_application(app_id=app_id)

    def _handle_signal(self, msg: SignalMessage) -> None:
        signal = Signal(msg.signal)
        self._logger.debug(
            f"SubroutineHandler at node {self.name} handles the signal {signal}"
        )
        if signal == Signal.STOP:
            self._logger.debug(f"SubroutineHandler at node {self.name} will stop")
            # Just mark that it will stop, to first send back the reply
            self._finished = True
        else:
            raise ValueError(f"Unkown signal {signal}")

    def _handle_open_epr_socket(
        self, msg: OpenEPRSocketMessage
    ) -> Generator[Any, None, None]:
        yield from self._executor.setup_epr_socket(
            epr_socket_id=msg.epr_socket_id,
            remote_node_id=msg.remote_node_id,
            remote_epr_socket_id=msg.remote_epr_socket_id,
        )



from netqasm.lang.encoding import RegisterName
from netqasm.lang.instr.vanilla import RotXInstruction, RotZInstruction
from netqasm.lang.ir import GenericInstr, ICmd, ProtoSubroutine
from netqasm.lang.operand import Register, Template
from netqasm.lang.parsing import parse_text_subroutine
from netqasm.lang.parsing.text import parse_text_protosubroutine
from netqasm.lang.subroutine import Subroutine
from netqasm.lang.version import NETQASM_VERSION


def test_protosubroutine_instantiation():
    commands = [
        ICmd(
            instruction=GenericInstr.ROT_X,
            operands=[
                Register(RegisterName.R, 0),
                Template("rot_num"),
                Template("rot_denom"),
            ],
        )
    ]
    arguments = ["rot_num", "rot_denom"]
    protosubrt = ProtoSubroutine(commands, arguments)
    print(protosubrt)

    protosubrt.instantiate(app_id=1, arguments={"rot_num": 3, "rot_denom": 1})
    print(protosubrt)


def test_subroutine_instantiation():
    instructions = [
        RotXInstruction(
            reg=Register(name=RegisterName.R, index=0),
            imm0=Template("rot_num"),
            imm1=Template("rot_denom"),
        ),
        RotXInstruction(
            reg=Register(name=RegisterName.R, index=0),
            imm0=Template("rot_num"),
            imm1=Template("rot_denom"),
        ),
    ]
    arguments = ["rot_num", "rot_denom"]
    subrt = Subroutine(instructions, arguments)
    print(subrt)

    subrt.instantiate(app_id=1, arguments={"rot_num": 3, "rot_denom": 1})
    print(subrt)


def test_protosubroutine_parsing():
    subroutine = """
rot_z R0 {num} {denom}
"""

    commands = [
        ICmd(
            instruction=GenericInstr.ROT_Z,
            operands=[
                Register(RegisterName.R, 0),
                Template("num"),
                Template("denom"),
            ],
        )
    ]
    arguments = ["num", "denom"]
    expected = ProtoSubroutine(commands, arguments)
    print(expected)

    protosubrt = parse_text_protosubroutine(subroutine)
    print(protosubrt)

    for cmd1, cmd2 in zip(protosubrt.commands, expected.commands):
        assert cmd1 == cmd2
    for arg1, arg2 in zip(protosubrt.arguments, expected.arguments):
        assert arg1 == arg2


def test_subroutine_parsing():
    subroutine = """
rot_z R0 {num} {denom}
"""

    instructions = [
        RotZInstruction(
            reg=Register(name=RegisterName.R, index=0),
            imm0=Template("num"),
            imm1=Template("denom"),
        ),
    ]
    arguments = ["num", "denom"]
    expected = Subroutine(instructions, arguments)

    subrt = parse_text_subroutine(subroutine)
    print(subrt)

    for instr1, instr2 in zip(subrt.instructions, expected.instructions):
        assert instr1 == instr2
    for arg1, arg2 in zip(subrt.arguments, expected.arguments):
        assert arg1 == arg2


def test_parse_and_instantiate():
    template = """
rot_z R0 {num} {denom}
"""

    expected_text = f"""
# NETQASM {NETQASM_VERSION[0]}.{NETQASM_VERSION[1]}
# APPID 0
rot_z R0 3 1
"""

    template: Subroutine = parse_text_subroutine(template)
    print(f"template: {template}")

    expected: Subroutine = parse_text_subroutine(expected_text)
    print(f"expected: {expected}")

    template.instantiate(app_id=0, arguments={"num": 3, "denom": 1})
    print(f"subroutine: {template}")

    assert template.instructions == expected.instructions


if __name__ == "__main__":
    test_protosubroutine_instantiation()
    test_subroutine_instantiation()
    test_subroutine_parsing()
    test_protosubroutine_parsing()
    test_parse_and_instantiate()



"""
Definitions of messages between the Host and the quantum node controller.
"""

import ctypes
from enum import Enum
from typing import Union

from netqasm.lang.encoding import INTEGER, Address, OptionalInt, Register
from netqasm.lang.subroutine import Subroutine

# C types for serialization
MESSAGE_TYPE = ctypes.c_uint8
MESSAGE_ID = ctypes.c_uint32
APP_ID = ctypes.c_uint32
NUM_QUBITS = ctypes.c_uint8
EPR_SOCKET_ID = INTEGER
EPR_FIDELITY = ctypes.c_uint8
NODE_ID = INTEGER
SIGNAL = ctypes.c_uint8

MESSAGE_TYPE_BYTES = len(bytes(MESSAGE_TYPE()))  # type: ignore


class MessageHeader(ctypes.Structure):
    _fields_ = [
        ("id", MESSAGE_ID),
        ("length", ctypes.c_uint32),
    ]

    @classmethod
    def len(cls):
        return len(bytes(cls()))

    def __str__(self):
        return f"{self.__class__.__name__}(id={self.id}, length={self.length})"


class MessageType(Enum):
    INIT_NEW_APP = 0x00
    OPEN_EPR_SOCKET = 0x01
    SUBROUTINE = 0x02
    STOP_APP = 0x03
    SIGNAL = 0x04


class Message(ctypes.Structure):
    _pack = 1
    _fields_ = [
        ("type", MESSAGE_TYPE),
    ]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        return cls.from_buffer_copy(raw)

    def __str__(self):
        to_print = f"{self.__class__.__name__}("
        for field_name, _ in self._fields_:
            to_print += f"{field_name}={getattr(self, field_name)}, "
        to_print = to_print[:-2] + ")"
        return to_print

    def __len__(self):
        return len(bytes(self))


class InitNewAppMessage(Message):
    """Message sent to the quantum node controller to register a new application."""

    _fields_ = [
        ("app_id", APP_ID),  # type: ignore
        ("max_qubits", NUM_QUBITS),
    ]

    TYPE = MessageType.INIT_NEW_APP

    def __init__(self, app_id=0, max_qubits=0):
        super().__init__(self.TYPE.value)
        self.app_id = app_id
        self.max_qubits = max_qubits


class OpenEPRSocketMessage(Message):
    """Message sent to the quantum node controller to open an EPR socket."""

    _fields_ = [
        ("app_id", APP_ID),  # type: ignore
        ("epr_socket_id", EPR_SOCKET_ID),  # type: ignore
        ("remote_node_id", NODE_ID),  # type: ignore
        ("remote_epr_socket_id", EPR_SOCKET_ID),  # type: ignore
        ("min_fidelity", EPR_FIDELITY),  # type: ignore
    ]

    TYPE = MessageType.OPEN_EPR_SOCKET

    def __init__(
        self,
        app_id=0,
        epr_socket_id=0,
        remote_node_id=0,
        remote_epr_socket_id=0,
        min_fidelity=100,
    ):
        super().__init__(self.TYPE.value)
        self.app_id = app_id
        self.epr_socket_id = epr_socket_id
        self.remote_node_id = remote_node_id
        self.remote_epr_socket_id = remote_epr_socket_id
        self.min_fidelity = min_fidelity


class SubroutineMessage:
    """Message sent to the quantum node controller to execute a subroutine."""

    TYPE = MessageType.SUBROUTINE

    def __init__(self, subroutine: Union[bytes, Subroutine]):
        """
        NOTE this message does not subclass from `Message` since it contains
        a subroutine which is defined separately and not as a `ctype` for now.
        Still this class defines the methods `__bytes__` and `deserialize_from`
        so that it can be packed and unpacked.
        The packed form of the message is:

        .. code-block:: text

            | TYP | SUBROUTINE ... |

        """
        self.type = self.TYPE.value
        if isinstance(subroutine, Subroutine):
            self.subroutine = bytes(subroutine)
        elif isinstance(subroutine, bytes):
            self.subroutine = subroutine
        else:
            raise TypeError(
                f"subroutine should be Subroutine or bytes, not {type(subroutine)}"
            )

    def __bytes__(self):
        return bytes(MESSAGE_TYPE(self.type)) + bytes(self.subroutine)

    def __len__(self):
        return len(bytes(self))

    @classmethod
    def deserialize_from(cls, raw: bytes):
        # NOTE we don't deserialize the subroutine here, since we need to know which flavour
        # is being used
        return cls(subroutine=raw[MESSAGE_TYPE_BYTES:])


class StopAppMessage(Message):
    """Message sent to the quantum node controller to stop/finish an application."""

    _fields_ = [
        ("app_id", APP_ID),  # type: ignore
    ]

    TYPE = MessageType.STOP_APP

    def __init__(self, app_id=0):
        super().__init__(self.TYPE.value)
        self.app_id = app_id


class Signal(Enum):
    STOP = 0


class SignalMessage(Message):
    """Message sent to the quantum node controller with a specific signal.

    Currently only used with the SquidASM simulator backend.
    """

    _fields_ = [
        ("signal", SIGNAL),
    ]

    TYPE = MessageType.SIGNAL

    def __init__(self, signal: Signal = Signal.STOP):
        super().__init__(self.TYPE.value)
        self.signal = signal.value


MESSAGE_CLASSES = {
    MessageType.INIT_NEW_APP: InitNewAppMessage,
    MessageType.OPEN_EPR_SOCKET: OpenEPRSocketMessage,
    MessageType.SUBROUTINE: SubroutineMessage,
    MessageType.STOP_APP: StopAppMessage,
    MessageType.SIGNAL: SignalMessage,
}


def deserialize_host_msg(raw: bytes) -> Message:
    """Convert a serialized message into a `Message` object

    :param raw: serialized message (string of bytes)
    :return: deserialized message object
    """
    message_type = MessageType(
        MESSAGE_TYPE.from_buffer_copy(raw[:MESSAGE_TYPE_BYTES]).value
    )
    message_class = MESSAGE_CLASSES[message_type]
    return message_class.deserialize_from(raw)  # type: ignore


class ReturnMessage(Message):
    """Base class for messages from the quantum node controller to the Host."""

    pass


class ReturnMessageType(Enum):
    DONE = 0x00
    ERR = 0x01
    RET_ARR = 0x02
    RET_REG = 0x03


class MsgDoneMessage(ReturnMessage):
    """Message to the Host that a subroutine has finished."""

    _fields_ = [
        ("msg_id", MESSAGE_ID),  # type: ignore
    ]

    TYPE = ReturnMessageType.DONE

    def __init__(self, msg_id=0):
        super().__init__(self.TYPE.value)
        self.msg_id = msg_id


class ErrorCode(Enum):
    GENERAL = 0x00
    NO_QUBIT = 0x01
    UNSUPP = 0x02


class ErrorMessage(ReturnMessage):
    """Message to the Host that an error occurred at the quantum node controller."""

    _fields_ = [
        ("err_code", ctypes.c_uint8),
    ]

    TYPE = ReturnMessageType.ERR

    def __init__(self, err_code):
        super().__init__(self.TYPE.value)
        self.err_code = err_code.value


class ReturnArrayMessageHeader(ctypes.Structure):
    """Header for a message with a returned array coming from the quantum node
    controller.
    """

    _pack = 1
    _fields_ = [
        ("address", Address),
        ("length", INTEGER),
    ]

    @classmethod
    def len(cls):
        return len(bytes(cls()))


class ReturnArrayMessage:
    """Message with a returned array coming from the quantum node controller."""

    TYPE = ReturnMessageType.RET_ARR

    def __init__(self, address, values):
        """NOTE this message does not subclass from `ReturnMessage` since
        the values is of variable length.
        Still this class defines the methods `__bytes__` and `deserialize_from`
        so that it can be packed and unpacked.

        The packed form of the message is:

        .. code-block:: text

            | ADDRESS | LENGTH | VALUES ... |

        """
        self.type = self.TYPE.value
        self.address = address
        self.values = values

    def __bytes__(self):
        array_type = OptionalInt * len(self.values)
        payload = array_type(*(OptionalInt(v) for v in self.values))
        hdr = ReturnArrayMessageHeader(
            address=Address(self.address),
            length=len(self.values),
        )
        return bytes(MESSAGE_TYPE(self.type)) + bytes(hdr) + bytes(payload)

    def __str__(self):
        return (
            f"{self.__class__.__name__}(address={self.address}, values={self.values})"
        )

    def __len__(self):
        return len(bytes(self))

    @classmethod
    def deserialize_from(cls, raw: bytes):
        raw = raw[MESSAGE_TYPE_BYTES:]
        hdr = ReturnArrayMessageHeader.from_buffer_copy(raw)
        array_type = OptionalInt * hdr.length
        raw = raw[ReturnArrayMessageHeader.len() :]
        values = list(v.value for v in array_type.from_buffer_copy(raw))
        return cls(address=hdr.address.address, values=values)


class ReturnRegMessage(ReturnMessage):
    """Message with a returned register coming from the quantum node controller."""

    _fields_ = [
        ("register", Register),  # type: ignore
        ("value", INTEGER),  # type: ignore
    ]

    TYPE = ReturnMessageType.RET_REG

    def __init__(self, register, value):
        super().__init__(self.TYPE.value)
        self.register = register
        self.value = value


RETURN_MESSAGE_CLASSES = {
    ReturnMessageType.DONE: MsgDoneMessage,
    ReturnMessageType.ERR: ErrorMessage,
    ReturnMessageType.RET_REG: ReturnRegMessage,
    ReturnMessageType.RET_ARR: ReturnArrayMessage,
}


def deserialize_return_msg(raw: bytes) -> Message:
    """Convert a serialized 'return' message into a `Message` object

    :param raw: serialized message (string of bytes)
    :return: deserialized message object
    """
    message_type = ReturnMessageType(
        MESSAGE_TYPE.from_buffer_copy(raw[:MESSAGE_TYPE_BYTES]).value
    )
    message_class = RETURN_MESSAGE_CLASSES[message_type]
    return message_class.deserialize_from(raw)  # type: ignore



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class LogConfig:
    track_lines: bool = False
    log_subroutines_dir: Optional[str] = None
    comm_log_dir: Optional[str] = None
    app_dir: Optional[str] = None
    lib_dirs: Optional[List[str]] = None
    log_dir: Optional[str] = None
    split_runs: bool = True



from .classical_communication import ThreadBroadcastChannel, ThreadSocket
from .epr_socket import EPRSocket
from .qubit import Qubit
from .toolbox import create_ghz, parity_meas, set_qubit_state, t_inverse, toffoli_gate



from collections import namedtuple
from random import randint

from netqasm.sdk import EPRSocket
from netqasm.sdk.external import BroadcastChannel, Socket
from netqasm.sdk.toolbox import create_ghz


def classical_anonymous_transmission(
    conn,
    sockets,
    num_nodes,
    sender=False,
    value=None,
):
    if sender:
        assert isinstance(value, bool), f"Value should be boolen, not {type(value)}"

    # Create a GHZ state
    q, _ = create_ghz(
        down_epr_socket=sockets.down_epr_socket,
        up_epr_socket=sockets.up_epr_socket,
        down_socket=sockets.down_socket,
        up_socket=sockets.up_socket,
        do_corrections=True,
    )

    # If sender and value is 1/True do Z flip
    if sender and value:
        q.Z()

    # Hadamard and measure
    q.H()
    m = q.measure()

    # Flush the commands to get the outcome
    conn.flush()

    # Send outcome to all other nodes
    broadcast_channel = sockets.broadcast_channel
    broadcast_channel.send(str(m))

    # Get measurements from all other nodes
    k = m
    for _ in range(num_nodes - 1):
        remote_node, m = broadcast_channel.recv()
        k += int(m)

    message = k % 2 == 1

    return message


def anonymous_epr(
    conn,
    sockets,
    num_nodes,
    sender=False,
    receiver=False,
):
    if sender and receiver:
        raise ValueError("Cannot be both sender and receiver")
    # Create a GHZ state
    q, _ = create_ghz(
        down_epr_socket=sockets.down_epr_socket,
        up_epr_socket=sockets.up_epr_socket,
        down_socket=sockets.down_socket,
        up_socket=sockets.up_socket,
        do_corrections=True,
    )

    # Get the broadcast_channel
    broadcast_channel = sockets.broadcast_channel

    if sender or receiver:
        b = randint(0, 1)
        broadcast_channel.send(str(b))
        if sender:
            if b == 1:
                q.Z()
    else:
        # Apply Hadamard
        q.H()
        # Measure qubit
        m = q.measure()

        conn.flush()

        # Broadcast outcome
        broadcast_channel.send(str(m))

        q = None

    # Receive all messages
    k = 0
    for _ in range(num_nodes - 1):
        _, m = broadcast_channel.recv()
        k += int(m)
    # Receiver does correction
    if receiver and k % 2 == 1:
        q.Z()

    return q


def quantum_anonymous_tranmission(
    conn,
    sockets,
    num_nodes,
    sender=False,
    receiver=False,
    qubit=None,
):
    if sender:
        if qubit is None:
            raise TypeError("The sender needs to provide the qubit to send")

    epr = anonymous_epr(
        conn=conn,
        sockets=sockets,
        num_nodes=num_nodes,
        sender=sender,
        receiver=receiver,
    )
    if sender:
        # Perform teleportation
        qubit.cnot(epr)
        qubit.H()
        m1 = qubit.measure()
        m2 = epr.measure()

        conn.flush()
        to_send = [bool(m1), bool(m2)]
    else:
        # Nothing to send
        to_send = [None, None]

    outcomes = []
    for m in to_send:
        outcome = classical_anonymous_transmission(
            conn=conn,
            sockets=sockets,
            num_nodes=num_nodes,
            sender=sender,
            value=m,
        )
        # outcomes.append(int(outcome))
        outcomes.append(outcome)

    if receiver:
        # Apply corrections and return qubit
        m1, m2 = outcomes
        if m2 == 1:
            epr.X()
        if m1 == 1:
            epr.Z()
        return epr
    else:
        return None


###################
# Setup functions #
###################
Sockets = namedtuple(
    "Sockets",
    [
        "broadcast_channel",
        "down_epr_socket",
        "down_socket",
        "up_epr_socket",
        "up_socket",
        "epr_sockets",
    ],
)


def setup_sockets(app_name, apps, log_config):
    broadcast_channel = _setup_broadcast_channel(app_name, apps, log_config)
    down_epr_socket, down_socket = _setup_down_sockets(app_name, apps, log_config)
    up_epr_socket, up_socket = _setup_up_sockets(app_name, apps, log_config)
    epr_sockets = [
        epr_socket
        for epr_socket in [down_epr_socket, up_epr_socket]
        if epr_socket is not None
    ]

    return Sockets(
        broadcast_channel=broadcast_channel,
        down_epr_socket=down_epr_socket,
        down_socket=down_socket,
        up_epr_socket=up_epr_socket,
        up_socket=up_socket,
        epr_sockets=epr_sockets,
    )


def _setup_broadcast_channel(app_name, app_names, log_config):
    # Create a broadcast_channel to send classical information
    remote_app_names = [an for an in app_names if an != app_name]
    broadcast_channel = BroadcastChannel(
        app_name,
        remote_app_names=remote_app_names,
        log_config=log_config,
        # Use socket ID to not mixup with the other sockets
        socket_id=1,
    )

    return broadcast_channel


def _setup_down_sockets(app_name, app_names, log_config):
    index = app_names.index(app_name)
    if index > 0:
        down_node = app_names[index - 1]
    else:
        down_node = None
    return _setup_sockets(app_name, down_node, log_config)


def _setup_up_sockets(app_name, app_names, log_config):
    index = app_names.index(app_name)
    if index < len(app_names) - 1:
        up_node = app_names[index + 1]
    else:
        up_node = None
    return _setup_sockets(app_name, up_node, log_config)


def _setup_sockets(app_name, other_node, log_config):
    if other_node is None:
        return None, None
    epr_socket = EPRSocket(other_node)
    socket = Socket(app_name, other_node, log_config=log_config)
    return epr_socket, socket



import random

from netqasm.logging.glob import set_log_level
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.epr_socket import EPRSocket

DebugConnection.node_ids = {
    "Alice": 0,
    "Bob": 1,
}


def main(no_output=False):
    n = 10

    epr_socket = EPRSocket(remote_app_name="Bob")
    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
        bit_flips = alice.new_array(
            init_values=[random.randint(0, 1) for _ in range(n)]
        )
        basis_flips = alice.new_array(
            init_values=[random.randint(0, 1) for _ in range(n)]
        )
        outcomes = alice.new_array(n)

        with epr_socket.create_context(number=n, sequential=True) as (q, pair):
            with bit_flips.get_future_index(pair).if_eq(1):
                q.X()
            with basis_flips.get_future_index(pair).if_eq(1):
                q.H()
            outcome = outcomes.get_future_index(pair)
            q.measure(outcome)

    if no_output:
        # Third message since the first two are for init application
        # and opening the EPR socket
        print(f"binary:\n{alice.storage[2]}")


if __name__ == "__main__":
    set_log_level("INFO")
    main()



class NVEprCompiler:
    """Utility methods for NV-specific compilation of EPR-related operations"""

    @classmethod
    def get_max_time_for_fidelity(cls, min_fidelity: int) -> int:
        """Convert a minimum fidelity requirement value to a maximum
        EPR generation duration in microseconds."""

        # Fidelity should be between 0 and 100.
        assert 0 <= min_fidelity <= 100

        # f = 50 -> time = 55_000 us
        # f = 80 -> time = 28_000 us
        # f = 100 -> time = 10_000 us
        return 100_000 - min_fidelity * 900



from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()


def run_alice():
    with NetQASMConnection("Alice") as alice:
        half = 5
        num = half * 2

        outcomes = alice.new_array(num)
        even = alice.new_array(init_values=[0]).get_future_index(0)

        with alice.loop(num) as i:
            q = Qubit(alice)
            with even.if_eq(0):
                q.X()
            outcome = outcomes.get_future_index(i)
            q.measure(outcome)
            even.add(1, mod=2)

        alice.flush()
        assert len(outcomes) == num
        print(f"outcomes = {list(outcomes)}")
        expected = [1, 0] * half
        print(f"expected = {expected}")
        assert list(outcomes) == expected


def test_measure_loop_context():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
        ]
    )
    simulate_application(app_instance, use_app_config=False, enable_logging=False)



from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()

num = 10


def run_alice():
    epr_socket = EPRSocket("bob")
    with NetQASMConnection("alice", epr_sockets=[epr_socket]) as alice:

        outcomes = alice.new_array(num)

        def post_create(conn, q, pair):
            q.H()
            outcome = outcomes.get_future_index(pair)
            q.measure(outcome)

        epr_socket.create(number=num, post_routine=post_create, sequential=True)

    return list(outcomes)


def run_bob():
    epr_socket = EPRSocket("alice")
    with NetQASMConnection("bob", epr_sockets=[epr_socket]) as bob:

        outcomes = bob.new_array(num)

        def post_recv(conn, q, pair):
            q.H()
            outcome = outcomes.get_future_index(pair)
            q.measure(outcome)

        epr_socket.recv(number=num, post_routine=post_recv, sequential=True)

    return list(outcomes)


def test_post_epr():
    app_instance = default_app_instance(
        [
            ("alice", run_alice),
            ("bob", run_bob),
        ]
    )
    results = simulate_application(
        app_instance, use_app_config=False, enable_logging=False
    )[0]
    print(results)

    assert results["app_alice"] == results["app_bob"]


if __name__ == "__main__":
    for _ in range(100):
        test_post_epr()



"""BroadcastChannel implementation using ThreadSockets."""

from __future__ import annotations

from typing import TYPE_CHECKING, Type

from ..broadcast_channel import BroadcastChannelBySockets
from .socket import ThreadSocket

if TYPE_CHECKING:
    from .. import socket


class ThreadBroadcastChannel(BroadcastChannelBySockets):
    _socket_class: Type[socket.Socket] = ThreadSocket



import numpy as np

from netqasm.logging.output import get_new_app_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket


def measure_basis_0(bob, q):
    q.rot_Y(angle=-np.pi / 4)
    return q.measure()


def measure_basis_1(bob, q):
    q.rot_Y(angle=np.pi / 4)
    return q.measure()


def format_corrections(m1, m2):
    if (m1, m2) == (0, 0):
        return "No corrections"
    elif (m1, m2) == (0, 1):
        return "X"
    elif (m1, m2) == (1, 0):
        return "Z"
    elif (m1, m2) == (1, 1):
        return "X and Z"


def format_measurement_basis(y):
    if y == 0:
        return "Z-basis rotated around Y by pi / 4"
    elif y == 1:
        return "Z-basis rotated around Y by -pi / 4"


def game_won(x, y, a, b):
    if x == 1 and y == 1:
        if a != b:
            return "Alice and Bob won the game, since x * y = 1 and a ^ b = 1"
        else:
            return "Alice and Bob lost the game, since x * y = 1 and a ^ b = 0"
    else:
        if a == b:
            return "Alice and Bob won the game, since x * y = 0 and a ^ b = 0"
        else:
            return "Alice and Bob lost the game, since x * y = 0 and a ^ b = 1"


def main(app_config=None, y=0):
    if not (y == 0 or y == 1):
        raise ValueError(f"y should be 0 or 1, not {y}")

    app_logger = get_new_app_logger(
        app_name=app_config.app_name, log_config=app_config.log_config
    )

    # Only to enforce global order of operations for nice visualization,
    # NOT needed to obtain CHSH outcome correlations.
    socket_alice = Socket("bob", "alice", log_config=app_config.log_config)

    socket = Socket("bob", "repeater", log_config=app_config.log_config)
    epr_socket = EPRSocket("repeater")

    bob = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )

    with bob:
        # Wait for entanglement with Alice through repeater
        epr = epr_socket.recv_keep()[0]
        bob.flush()

        # Receive teleportation corrections
        msg = socket.recv()
        m1, m2 = eval(msg)
        app_logger.log(
            f"Applying teleportation corrections {format_corrections(m1, m2)}"
        )
        if m2 == 1:
            epr.X()
        if m1 == 1:
            epr.Z()
        bob.flush()

        # Tell Alice entanglement has been established
        socket_alice.send_silent("")

        # CHSH strategy: measure in one of 2 bases depending on y.
        app_logger.log(f"Measuring in basis y = {y}...")
        if y == 0:
            b = measure_basis_0(bob, epr)
        else:
            b = measure_basis_1(bob, epr)

    msg = socket_alice.recv_silent()
    x, a = eval(msg)

    app_logger.log(f"Bob outputs b = {b}")
    return {
        "b": int(b),
        "corrections": format_corrections(m1, m2),
        "basis": format_measurement_basis(y),
        "game_won": game_won(x, y, a, b),
    }


if __name__ == "__main__":
    main()
